from abc import ABCMeta, abstractmethod
from typing import Any

from .IStartStop import IStartStop


class IAutoGuiding(IStartStop, metaclass=ABCMeta):
    """The module can perform auto-guiding."""

    __module__ = "pyobs.interfaces"

    @abstractmethod
    async def set_exposure_time(self, exposure_time: float, **kwargs: Any) -> None:
        """Set the exposure time for the auto-guider.

        Args:
            exposure_time: Exposure time in secs.
        """
        ...


__all__ = ["IAutoGuiding"]



from __future__ import annotations
import logging
from typing import Union, Dict, Tuple, Optional, List, Any, TYPE_CHECKING

from pyobs.object import get_object
from pyobs.robotic.scripts import Script
from pyobs.robotic.task import Task
from pyobs.utils.logger import DuplicateFilter
from pyobs.utils.time import Time

if TYPE_CHECKING:
    from pyobs.robotic import TaskRunner, TaskSchedule, TaskArchive

log = logging.getLogger(__name__)

# logger for logging name of task
task_name_logger = logging.getLogger(__name__ + ":task_name")
task_name_logger.addFilter(DuplicateFilter())


class ConfigStatus:
    """Status of a single configuration."""

    def __init__(self, state: str = "ATTEMPTED", reason: str = ""):
        """Initializes a new Status with an ATTEMPTED."""
        self.start: Time = Time.now()
        self.end: Optional[Time] = None
        self.state: str = state
        self.reason: str = reason
        self.time_completed: float = 0.0

    def finish(
        self, state: Optional[str] = None, reason: Optional[str] = None, time_completed: float = 0.0
    ) -> "ConfigStatus":
        """Finish this status with the given values and the current time.

        Args:
            state: State of configuration
            reason: Reason for that state
            time_completed: Completed time [s]
        """
        if state is not None:
            self.state = state
        if reason is not None:
            self.reason = reason
        self.time_completed = time_completed
        self.end = Time.now()
        return self

    def to_json(self) -> Dict[str, Any]:
        """Convert status to JSON for sending to portal."""
        return {
            "state": self.state,
            "summary": {
                "state": self.state,
                "reason": self.reason,
                "start": self.start.isot,
                "end": "" if self.end is None else self.end.isot,
                "time_completed": self.time_completed,
            },
        }


class LcoTask(Task):
    """A task from the LCO portal."""

    def __init__(self, config: Dict[str, Any], **kwargs: Any):
        """Init LCO task (called request there).

        Args:
            config: Configuration for task
        """
        Task.__init__(self, **kwargs)

        # store stuff
        self.config = config
        self.cur_script: Optional[Script] = None

    @property
    def id(self) -> Any:
        """ID of task."""
        if "request" in self.config and "id" in self.config["request"]:
            return self.config["request"]["id"]
        else:
            raise ValueError("No id found in request.")

    @property
    def name(self) -> str:
        """Returns name of task."""
        if "name" in self.config and isinstance(self.config["name"], str):
            return self.config["name"]
        else:
            raise ValueError("No name found in request group.")

    @property
    def duration(self) -> float:
        """Returns estimated duration of task in seconds."""
        if (
            "request" in self.config
            and "duration" in self.config["request"]
            and isinstance(self.config["request"]["duration"], int)
        ):
            return float(self.config["request"]["duration"])
        else:
            raise ValueError("No duration found in request.")

    def __eq__(self, other: object) -> bool:
        """Compares to tasks."""
        if isinstance(other, LcoTask):
            return self.config == other.config
        else:
            return False

    @property
    def start(self) -> Time:
        """Start time for task"""
        if "start" in self.config and isinstance(self.config["start"], Time):
            return self.config["start"]
        else:
            raise ValueError("No start time found in request group.")

    @property
    def end(self) -> Time:
        """End time for task"""
        if "end" in self.config and isinstance(self.config["end"], Time):
            return self.config["end"]
        else:
            raise ValueError("No end time found in request group.")

    @property
    def observation_type(self) -> str:
        """Returns observation_type of this task.

        Returns:
            observation_type of this task.
        """
        if "observation_type" in self.config:
            return self.config["observation_type"]
        else:
            raise ValueError("No observation_type found in request group.")

    @property
    def can_start_late(self) -> bool:
        """Whether this tasks is allowed to start later than the user-set time, e.g. for flatfields.

        Returns:
            True, if task can start late.
        """
        return self.observation_type == "DIRECT"

    def _get_config_script(self, config: Dict[str, Any], scripts: Optional[Dict[str, Script]] = None) -> Script:
        """Get config script for given configuration.

        Args:
            config: Config to create runner for.

        Returns:
            Script for running config

        Raises:
            ValueError: If could not create runner.
        """

        # what do we run?
        config_type = config["type"]
        if scripts is None or config_type not in scripts:
            raise ValueError('No script found for configuration type "%s".' % config_type)

        # create script handler
        return get_object(
            scripts[config_type],
            Script,
            configuration=config,
            comm=self.comm,
            observer=self.observer,
        )

    async def can_run(self, scripts: Optional[Dict[str, Script]] = None) -> bool:
        """Checks, whether this task could run now.

        Returns:
            True, if task can run now.
        """

        # get logger for task name and log
        task_name_logger.info(f"Checking whether task {self.name} can run...")

        # loop configurations
        req = self.config["request"]
        for config in req["configurations"]:
            # get config runner
            runner = self._get_config_script(config, scripts)

            # if any runner can run, we proceed
            try:
                if await runner.can_run():
                    return True
            except Exception:
                log.exception("Error on evaluating whether task can run.")
                return False

        # no config found that could run
        return False

    async def run(
        self,
        task_runner: TaskRunner,
        task_schedule: Optional[TaskSchedule] = None,
        task_archive: Optional[TaskArchive] = None,
        scripts: Optional[Dict[str, Script]] = None,
    ) -> None:
        """Run a task"""
        from pyobs.robotic.lco import LcoTaskSchedule

        # get request
        req = self.config["request"]

        # loop configurations
        status: Optional[ConfigStatus]
        for config in req["configurations"]:
            # send an ATTEMPTED status
            if isinstance(task_schedule, LcoTaskSchedule):
                status = ConfigStatus()
                self.config["state"] = "ATTEMPTED"
                await task_schedule.send_update(config["configuration_status"], status.finish().to_json())

            # get config runner
            script = self._get_config_script(config, scripts)

            # can run?
            if not await script.can_run():
                log.warning("Cannot run config.")
                continue

            # run config
            log.info("Running config...")
            self.cur_script = script
            status = await self._run_script(
                script, task_runner=task_runner, task_schedule=task_schedule, task_archive=task_archive
            )
            self.cur_script = None

            # send status
            if status is not None and isinstance(task_schedule, LcoTaskSchedule):
                self.config["state"] = status.state
                await task_schedule.send_update(config["configuration_status"], status.to_json())

        # finished task
        log.info("Finished task.")

    async def _run_script(
        self,
        script: Script,
        task_runner: TaskRunner,
        task_schedule: Optional[TaskSchedule] = None,
        task_archive: Optional[TaskArchive] = None,
    ) -> Union[ConfigStatus, None]:
        """Run a config

        Args:
            script: Script to run

        Returns:
            Configuration status to send to portal
        """

        # at least we tried...
        config_status = ConfigStatus()

        try:
            # run it
            log.info("Running task %d: %s...", self.id, self.config["name"])
            await script.run(task_runner=task_runner, task_schedule=task_schedule, task_archive=task_archive)

            # finished config
            config_status.finish(state="COMPLETED", time_completed=script.exptime_done)

        except InterruptedError:
            log.warning("Task execution was interrupted.")
            config_status.finish(
                state="FAILED", reason="Task execution was interrupted.", time_completed=script.exptime_done
            )

        except Exception:
            log.exception("Something went wrong.")
            config_status.finish(state="FAILED", reason="Something went wrong", time_completed=script.exptime_done)

        # finished
        return config_status

    def is_finished(self) -> bool:
        """Whether task is finished."""
        if "config" in self.config and isinstance(self.config["state"], str):
            return self.config["state"] != "PENDING"
        else:
            return False

    def get_fits_headers(self, namespaces: Optional[List[str]] = None) -> Dict[str, Tuple[Any, str]]:
        """Returns FITS header for the current status of this module.

        Args:
            namespaces: If given, only return FITS headers for the given namespaces.

        Returns:
            Dictionary containing FITS headers.
        """

        # get header from script
        hdr = self.cur_script.get_fits_headers(namespaces) if self.cur_script is not None else {}

        # return it
        return hdr


__all__ = ["LcoTask"]



import asyncio
import logging
import os
import subprocess
from typing import Optional, Any

from pyobs.modules import Module
from pyobs.interfaces import IAutonomous

log = logging.getLogger(__name__)


class AutonomousWarning(Module):
    """A module that can plays a warning sound while an IAutonomous module is running."""

    __module__ = "pyobs.modules.utils"

    def __init__(
        self,
        warn_sound: str,
        warn_interval: float = 1,
        start_sound: Optional[str] = None,
        started_sound: Optional[str] = None,
        stop_sound: Optional[str] = None,
        stopped_sound: Optional[str] = None,
        player: str = "mpg123",
        trigger_file: Optional[str] = None,
        **kwargs: Any,
    ):
        """Initialize a new warning.

        Args:
            warn_sound: Name of file to play.
            warn_interval: Interval in seconds between sounds.
            start_sound: Sound to play when starting systems.
            started_sound: Sound to play when systems started.
            stop_sound: Sound to play when stopping systems.
            stopped_sound: Sound to play when systems stopped.
            trigger_file: File, which triggers to switch on-off and vice versa, when created.
                Will be deleted afterwards.
        """
        Module.__init__(self, **kwargs)

        # store
        self._warn_sound = warn_sound
        self._warn_interval = warn_interval
        self._start_sound = start_sound
        self._started_sound = started_sound
        self._stop_sound = stop_sound
        self._stopped_sound = stopped_sound
        self._trigger_file = trigger_file
        self._player = player
        self._autonomous = False

        # threads
        self.add_background_task(self._heartbeat)
        self.add_background_task(self._check_autonomous)
        self.add_background_task(self._check_trigger)

    def _play_sound(self, sound: str) -> None:
        """Play a sound.

        Args:
            sound: Sound file to play.
        """

        # no sound given?
        if sound is None:
            return

        # play sound and set stderr to devnull to avoid warning
        subprocess.Popen(
            [self._player, "-q", sound],
            stderr=subprocess.DEVNULL,
        ).wait()

    async def _heartbeat(self) -> None:
        """Play sound in given interval, if an autonomous module is running."""

        while True:
            # play sound
            if self._autonomous:
                self._play_sound(self._warn_sound)

            # sleep
            await asyncio.sleep(self._warn_interval)

    async def _check_autonomous(self) -> None:
        """Checks for autonomous modules."""

        while True:
            # check for autonomous modules
            autonomous = list(await self.comm.clients_with_interface(IAutonomous))
            is_auto = any([await (await self.comm.proxy(a, IAutonomous)).is_running().wait() for a in autonomous])

            # did it change?
            if is_auto != self._autonomous:
                log.info("Robotic systems %s.", "started" if is_auto else "stopped")
                if self._stop_sound is not None and is_auto:
                    self._play_sound(self._stop_sound)
                elif self._start_sound is not None and not is_auto:
                    self._play_sound(self._start_sound)

            # store it
            self._autonomous = is_auto

            # sleep a little
            await asyncio.sleep(1)

    async def _check_trigger(self) -> None:
        """Checks for trigger to start/stop autonomous modules."""

        while True:
            # does file exist?
            if self._trigger_file is not None and os.path.exists(self._trigger_file):
                # check for autonomous modules
                autonomous = list(await self.comm.clients_with_interface(IAutonomous))
                is_auto = any([await (await self.proxy(a, IAutonomous)).is_running() for a in autonomous])

                # play sound
                if self._stop_sound is not None and is_auto:
                    self._play_sound(self._stop_sound)
                elif self._start_sound is not None and not is_auto:
                    self._play_sound(self._start_sound)

                # loop all modules
                log.info("%s robotic systems:", "Stopping" if is_auto else "Starting")
                for auto in autonomous:
                    # get proxy
                    log.info("  - %s", auto)
                    proxy = await self.comm.proxy(auto)

                    # start/stop
                    if is_auto:
                        await proxy.stop()
                    else:
                        await proxy.start()
                log.info("Finished.")

                # remove file
                os.remove(self._trigger_file)

            # sleep a little
            await asyncio.sleep(1)


__all__ = ["AutonomousWarning"]



import argparse
import glob
import logging
import os
import subprocess
import sys
import time
from typing import Optional, List

import yaml

log = logging.getLogger("pyobs")


class PyobsDaemon(object):
    def __init__(
        self,
        config_path: str,
        run_path: str,
        log_path: str,
        log_level: str = "info",
        chuid: Optional[str] = None,
        start_stop_daemon: str = "start-stop-daemon",
    ):
        self._config_path = config_path
        self._run_path = run_path
        self._log_path = log_path
        self._log_level = log_level
        self._chuid = chuid
        self._start_stop_daemon = start_stop_daemon

        # find pyobs executable
        filenames = [
            os.path.join(os.path.dirname(os.path.abspath(sys.argv[0])), "pyobs"),
            "/usr/bin/pyobs",
            "/usr/local/bin/pyobs",
        ]
        for filename in filenames:
            if os.path.exists(filename):
                self._pyobs_exec = filename
                break
        else:
            raise ValueError("Could not find pyobs executable.")

        # get configs and running
        self._configs = self._get_configs()
        self._running = self._get_running()

    def _get_configs(self) -> List[str]:
        # get configuration files, ignore those ending on .shared.yaml
        tmp = sorted(glob.glob(os.path.join(self._config_path, "*.yaml")))
        return list(filter(lambda t: not t.endswith(".shared.yaml"), tmp))

    def _get_running(self) -> List[str]:
        # get PID files
        pid_files = sorted(glob.glob(os.path.join(self._run_path, "*.pid")))

        # loop files
        running = []
        for pid_file in pid_files:
            # get pid
            pid = self._pid(self._service(pid_file))
            if pid is None:
                print("No PID file found.")
                continue

            # check for running
            try:
                os.kill(pid, 0)
            except OSError:
                print("Removing PID file %s without process..." % os.path.basename(pid_file))
                os.remove(pid_file)
            else:
                running.append(pid_file)

        # return running processes
        return running

    def start(self, services: Optional[List[str]] = None) -> None:
        # get list of running processes
        running = [self._service(r) for r in self._running]
        configs = [self._service(r) for r in self._configs]

        # if no services are given, start all
        if services is None:
            # ignore all configs that start with an underscore, those need to be started explicitly
            services = [self._service(c) for c in configs if not os.path.basename(c).startswith("_")]

        # loop configs
        for service in sorted(services):
            # exists?
            if service not in configs:
                print("Service %s does not exists." % service)
                sys.exit(1)

            # start it?
            if service in running:
                print("%s already running." % service)
            else:
                print("Starting %s..." % service)
                self._start_service(service)

    def stop(self, services: Optional[List[str]] = None) -> None:
        # if no services are given, stop all
        if services is None:
            services = [self._service(r) for r in self._running]

        # loop running and stop them
        for service in services:
            print("Stopping %s..." % service)
            self._stop_service(service)

    def restart(self, services: Optional[List[str]] = None) -> None:
        # stop all services
        self.stop(services=services)

        # sleep a little and get running
        time.sleep(1)
        self._running = self._get_running()

        # start all services
        self.start(services=services)

    def status(self, services: Optional[List[str]] = None) -> None:
        # get all configs and running
        configs = [self._service(r) for r in self._configs]
        running = [self._service(r) for r in self._running]

        # if no services are given, get all
        if services is None:
            services = sorted(list(set(configs + running)))

        # print them
        print("cfg run service")
        for p in services:
            print(("[X]" if p in configs else "[ ]") + " " + ("[X]" if p in running else "[ ]") + " " + p)

    def _start_service(self, service: str) -> None:
        # get PID file
        pid_file = self._pid_file(service)

        # define command
        cmd = []
        cmd.extend([self._start_stop_daemon, "--start", "--quiet", "--pidfile", pid_file])

        # change user?
        if self._chuid:
            cmd.extend(["--chuid", self._chuid])

        # call to pyobs
        cmd.extend(
            [
                "--exec",
                self._pyobs_exec,
                "--",
                "--pid-file",
                pid_file,
                "--log-file",
                self._log_file(service),
                "--log-level",
                self._log_level,
                self._config_file(service),
            ]
        )

        # execute
        subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    def _stop_service(self, service: str) -> None:
        # get service name and PID
        pid_file = self._pid_file(service)

        # stop service
        cmd = [self._start_stop_daemon, "--stop", "--quiet", "--oknodo", "--pidfile", pid_file]
        if self._chuid:
            cmd.extend(["--user", self._chuid[: self._chuid.find(":")]])
        subprocess.call(cmd)

    @staticmethod
    def _service(config_file: str) -> str:
        # get basename without extension
        return os.path.splitext(os.path.basename(config_file))[0]

    def _config_file(self, service: str) -> str:
        # get pid file
        return os.path.join(self._config_path, service + ".yaml")

    def _pid_file(self, service: str) -> str:
        # get pid file
        return os.path.join(self._run_path, service + ".pid")

    def _log_file(self, service: str) -> str:
        # get pid file
        return os.path.join(self._log_path, service + ".log")

    def _pid(self, service: str) -> Optional[int]:
        # get pid file
        pid_file = self._pid_file(service)
        if not os.path.exists(pid_file):
            return None

        # get pid
        with open(pid_file, "r") as f:
            return int(f.read())


def main() -> None:
    # try to load config file
    config_filename = os.path.expanduser("~/.pyobs/pyobsd.yaml")
    config = {}
    if os.path.exists(config_filename):
        with open(config_filename, "r") as f:
            config = yaml.safe_load(f)

    # init parser
    parser = argparse.ArgumentParser(description="Daemon for pyobs")
    parser.add_argument("-p", "--path", type=str, default=config.get("path", "/opt/pyobs"))
    parser.add_argument("-c", "--config-path", type=str, default=config.get("config-path", "config"))
    parser.add_argument("-r", "--run-path", type=str, default=config.get("run-path", "run"))
    parser.add_argument("-l", "--log-path", type=str, default=config.get("log-path", "log"))
    parser.add_argument(
        "--log-level",
        type=str,
        choices=["critical", "error", "warning", "info", "debug"],
        default=config.get("log-level", "info"),
    )
    parser.add_argument("--chuid", type=str, default=config.get("chuid", "pyobs:pyobs"))
    parser.add_argument(
        "--start-stop-daemon", type=str, default=config.get("start-stop-daemon", "/sbin/start-stop-daemon")
    )
    parser.add_argument("command", type=str, choices=["start", "stop", "restart", "status"])
    parser.add_argument("services", type=str, nargs="*")
    args = parser.parse_args()

    # init daemon
    daemon = PyobsDaemon(
        os.path.join(args.path, args.config_path),
        os.path.join(args.path, args.run_path),
        os.path.join(args.path, args.log_path),
        log_level=args.log_level,
        chuid=args.chuid,
        start_stop_daemon=args.start_stop_daemon,
    )

    # run
    cmd = getattr(daemon, args.command)
    cmd(services=args.services if args.services else None)


if __name__ == "__main__":
    main()



from typing import Dict, Tuple


class SkyflatPriorities:
    """Base class for sky flat priorities."""

    __module__ = "pyobs.utils.skyflats.priorities"

    async def __call__(self) -> Dict[Tuple[str, Tuple[int, int]], float]:
        return {}


__all__ = ["SkyflatPriorities"]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyobs',
 'pyobs.cli',
 'pyobs.comm',
 'pyobs.comm.dummy',
 'pyobs.comm.sleekxmpp',
 'pyobs.comm.sleekxmpp.xep_0009',
 'pyobs.comm.sleekxmpp.xep_0009_timeout',
 'pyobs.comm.sleekxmpp.xep_0009_timeout.stanza',
 'pyobs.comm.slixmpp',
 'pyobs.comm.slixmpp.xep_0009',
 'pyobs.comm.slixmpp.xep_0009_timeout',
 'pyobs.comm.slixmpp.xep_0009_timeout.stanza',
 'pyobs.events',
 'pyobs.images',
 'pyobs.images.meta',
 'pyobs.images.processors',
 'pyobs.images.processors.astrometry',
 'pyobs.images.processors.detection',
 'pyobs.images.processors.exptime',
 'pyobs.images.processors.misc',
 'pyobs.images.processors.offsets',
 'pyobs.images.processors.photometry',
 'pyobs.interfaces',
 'pyobs.mixins',
 'pyobs.modules',
 'pyobs.modules.camera',
 'pyobs.modules.flatfield',
 'pyobs.modules.focus',
 'pyobs.modules.image',
 'pyobs.modules.pointing',
 'pyobs.modules.robotic',
 'pyobs.modules.roof',
 'pyobs.modules.telescope',
 'pyobs.modules.test',
 'pyobs.modules.utils',
 'pyobs.modules.weather',
 'pyobs.robotic',
 'pyobs.robotic.lco',
 'pyobs.robotic.lco.scripts',
 'pyobs.robotic.scripts',
 'pyobs.utils',
 'pyobs.utils.archive',
 'pyobs.utils.focusseries',
 'pyobs.utils.offsets',
 'pyobs.utils.pipeline',
 'pyobs.utils.publisher',
 'pyobs.utils.simulation',
 'pyobs.utils.skyflats',
 'pyobs.utils.skyflats.pointing',
 'pyobs.utils.skyflats.priorities',
 'pyobs.utils.threads',
 'pyobs.vfs',
 'pyobs.vfs.filelists']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'aiohttp>=3.8.1,<4.0.0',
 'astroplan>=0.8,<0.9',
 'astropy>=5.0,<6.0',
 'astroquery>=0.4.4,<0.5.0',
 'ccdproc>=2.2,<3.0',
 'lmfit>=1.0,<2.0',
 'numpy>=1.21,<2.0',
 'pandas>=1.1.5,<2.0.0',
 'paramiko>=2.8,<3.0',
 'photutils>=1.2,<2.0',
 'py-expression-eval>=0.3,<0.4',
 'python-telegram-bot>=13.8,<14.0',
 'pytz>=2021.3,<2022.0',
 'requests>=2.26,<3.0',
 'scipy>=1.7.2,<2.0.0',
 'single-source>=0.2.0,<0.3.0',
 'slixmpp>=1.7.1,<2.0.0',
 'tornado>=6.1,<7.0',
 'typing-extensions>=4.0,<5.0']

extras_require = \
{':sys_platform == "linux"': ['sep>=1.2,<2.0',
                              'python-daemon>=2.3,<3.0',
                              'asyncinotify>=2.0.2,<3.0.0']}

entry_points = \
{'console_scripts': ['pyobs = pyobs.cli.pyobs:main',
                     'pyobsd = pyobs.cli.pyobsd:main',
                     'pyobsw = pyobs.cli.pyobsw:main']}

setup_kwargs = {
    'name': 'pyobs-core',
    'version': '0.18.15',
    'description': 'robotic telescope software',
    'long_description': None,
    'author': 'Tim-Oliver Husser',
    'author_email': 'thusser@uni-goettingen.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<3.11',
}


setup(**setup_kwargs)



import logging
from slixmpp.plugins.base import BasePlugin
from slixmpp.plugins.xep_0009.stanza.RPC import RPCQuery
from slixmpp.xmlstream.stanzabase import register_stanza_plugin
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import MatchXPath

from . import stanza


log = logging.getLogger(__name__)


class XEP_0009_timeout(BasePlugin):
    """A plugin for SleekXMPP, adding a timeout to RPC calls."""

    name = "xep_0009_timeout"
    description = "XEP-0009-timeout: Jabber-RPC timeout extension"
    dependencies = {"xep_0009"}
    stanza = stanza

    def plugin_init(self):
        register_stanza_plugin(RPCQuery, stanza.MethodTimeout)

        self.xmpp.register_handler(
            Callback(
                "RPC Call",
                MatchXPath(
                    "{%s}iq/{%s}query/{%s}methodTimeout"
                    % (self.xmpp.default_ns, RPCQuery.namespace, RPCQuery.namespace)
                ),
                self._handle_method_timeout,
            )
        )

        self.xmpp.add_event_handler("jabber_rpc_method_timeout", self._on_jabber_rpc_method_timeout)

    def make_iq_method_timeout(self, pid, pto, timeout):
        iq = self.xmpp.make_iq_result(pid)
        iq["to"] = pto
        iq["from"] = self.xmpp.boundjid.full
        iq.enable("rpc_query")
        iq["rpc_query"]["method_timeout"]["timeout"] = timeout
        return iq

    def _handle_method_timeout(self, iq):
        log.debug("Incoming Jabber-RPC timeout from %s", iq["from"])
        self.xmpp.event("jabber_rpc_method_timeout", iq)

    def _on_jabber_rpc_method_timeout(self, iq, forwarded=False):
        """A default handler for Jabber-RPC method timeout. If another handler is registered,
        this one will defer and not run.

        If this handler is called by your own custom handler with forwarded set to True, then it will run as normal.
        """
        if not forwarded and self.xmpp.event_handled("jabber_rpc_method_timeout") > 1:
            return

        # Reply with error by default
        error = self.xmpp["xep_0009"]._item_not_found(iq)
        error.send()



import asyncio
import logging
from typing import Union, List, Dict, Tuple, Any, Optional, cast
import astropy.units as u

from pyobs.modules import Module
from pyobs.events.taskfinished import TaskFinishedEvent
from pyobs.events.taskstarted import TaskStartedEvent
from pyobs.interfaces import IFitsHeaderBefore, IAutonomous
from pyobs.robotic.task import Task
from pyobs.utils.time import Time
from pyobs.robotic import TaskRunner, TaskSchedule

log = logging.getLogger(__name__)


class Mastermind(Module, IAutonomous, IFitsHeaderBefore):
    """Mastermind for a full robotic mode."""

    __module__ = "pyobs.modules.robotic"

    def __init__(
        self,
        schedule: Union[TaskSchedule, Dict[str, Any]],
        runner: Union[TaskRunner, Dict[str, Any]],
        allowed_late_start: int = 300,
        allowed_overrun: int = 300,
        **kwargs: Any,
    ):
        """Initialize a new auto focus system.

        Args:
            schedule: Object that can return schedule.
            allowed_late_start: Allowed seconds to start late.
            allowed_overrun: Allowed time for a task to exceed it's window in seconds
        """
        Module.__init__(self, **kwargs)

        # store
        self._allowed_late_start = allowed_late_start
        self._allowed_overrun = allowed_overrun
        self._running = False

        # add thread func
        self.add_background_task(self._run_thread, True)

        # get schedule and runner
        self._task_schedule = self.add_child_object(schedule, TaskSchedule)
        self._task_runner = self.add_child_object(runner, TaskRunner)

        # observation name and exposure number
        self._task = None
        self._obs = None
        self._exp = None

    async def open(self) -> None:
        """Open module."""
        await Module.open(self)

        # subscribe to events
        if self.comm:
            await self.comm.register_event(TaskStartedEvent)
            await self.comm.register_event(TaskFinishedEvent)

        # start
        self._running = True

    async def start(self, **kwargs: Any) -> None:
        """Starts a service."""
        log.info("Starting robotic system...")
        self._running = True

    async def stop(self, **kwargs: Any) -> None:
        """Stops a service."""
        log.info("Stopping robotic system...")
        self._running = False

    async def is_running(self, **kwargs: Any) -> bool:
        """Whether a service is running."""
        return self._running

    async def _run_thread(self) -> None:
        # wait a little
        await asyncio.sleep(1)

        # flags
        first_late_start_warning = True

        # run until closed
        while True:
            # not running?
            if not self._running:
                # sleep a little and continue
                await asyncio.sleep(1)
                continue

            # get now
            now = Time.now()

            # find task that we want to run now
            task: Optional[Task] = self._task_schedule.get_task(now)
            if task is None or not await self._task_runner.can_run(task):
                # no task found
                await asyncio.sleep(10)
                continue

            # starting too late?
            if not task.can_start_late:
                late_start = now - task.start
                if late_start > self._allowed_late_start * u.second:
                    # only warn once
                    if first_late_start_warning:
                        log.warning(
                            "Time since start of window (%.1f) too long (>%.1f), skipping task...",
                            late_start.to_value("second"),
                            self._allowed_late_start,
                        )
                    first_late_start_warning = False

                    # sleep a little and skip
                    await asyncio.sleep(10)
                    continue

            # reset warning
            first_late_start_warning = True

            # task is definitely not None here
            self._task = cast(Task, task)

            # ETA
            eta = now + self._task.duration * u.second

            # send event
            await self.comm.send_event(TaskStartedEvent(name=self._task.name, id=self._task.id, eta=eta))

            # run task in thread
            log.info("Running task %s...", self._task.name)
            await self._task_runner.run_task(self._task, task_schedule=self._task_schedule)

            # send event
            await self.comm.send_event(TaskFinishedEvent(name=self._task.name, id=self._task.id))

            # finish
            log.info("Finished task %s.", self._task.name)
            self._task = None

    async def get_fits_header_before(
        self, namespaces: Optional[List[str]] = None, **kwargs: Any
    ) -> Dict[str, Tuple[Any, str]]:
        """Returns FITS header for the current status of this module.

        Args:
            namespaces: If given, only return FITS headers for the given namespaces.

        Returns:
            Dictionary containing FITS headers.
        """

        # inside an observation?
        if self._task is not None:
            hdr = self._task.get_fits_headers()
            hdr["TASK"] = self._task.name, "Name of task"
            hdr["REQNUM"] = str(self._task.id), "Unique ID of task"
            return hdr
        else:
            return {}


__all__ = ["Mastermind"]



import logging
from typing import Any

from pyobs.modules import Module
from pyobs.interfaces import IAutoGuiding

log = logging.getLogger(__name__)


class DummyAutoGuiding(Module, IAutoGuiding):
    """An auto-guiding system."""

    __module__ = "pyobs.modules.guiding"

    def __init__(self, **kwargs: Any):
        Module.__init__(self, **kwargs)
        self._running = False

    async def set_exposure_time(self, exposure_time: float, **kwargs: Any) -> None:
        pass

    async def start(self, **kwargs: Any) -> None:
        log.info("Start guiding.")
        self._running = True

    async def stop(self, **kwargs: Any) -> None:
        log.info("Stop guiding.")
        self._running = False

    async def is_running(self, **kwargs: Any) -> bool:
        return self._running


__all__ = ["DummyAutoGuiding"]



from abc import ABCMeta, abstractmethod
from typing import Tuple, Any

from .interface import Interface


class IPointingAltAz(Interface, metaclass=ABCMeta):
    """The module can move to Alt/Az coordinates, usually combined with :class:`~pyobs.interfaces.ITelescope`."""

    __module__ = "pyobs.interfaces"

    @abstractmethod
    async def move_altaz(self, alt: float, az: float, **kwargs: Any) -> None:
        """Moves to given coordinates.

        Args:
            alt: Alt in deg to move to.
            az: Az in deg to move to.

        Raises:
            MoveError: If device could not be moved.
        """
        ...

    @abstractmethod
    async def get_altaz(self, **kwargs: Any) -> Tuple[float, float]:
        """Returns current Alt and Az.

        Returns:
            Tuple of current Alt and Az in degrees.
        """
        ...


__all__ = ["IPointingAltAz"]



from __future__ import annotations

from astropy.time import Time
from typing import Union, Optional, TYPE_CHECKING, Dict, Any

from pyobs.object import create_object, Object

if TYPE_CHECKING:
    from pyobs.utils.simulation.telescope import SimTelescope
    from pyobs.utils.simulation.camera import SimCamera


class SimWorld(Object):
    """A simulated world."""

    __module__ = "pyobs.utils.simulation"

    def __init__(
        self,
        time: Optional[Union[Time, str]] = None,
        telescope: Optional[Union["SimTelescope", Dict[str, Any]]] = None,
        camera: Optional[Union["SimCamera", Dict[str, Any]]] = None,
        **kwargs: Any,
    ):
        """Initializes a new simulated world.

        Args:
            time: Time at start of simulation.
            telescope: Telescope to use.
            camera: Camera to use.
            observer: Observer to use.
            *args:
            **kwargs:
        """
        from .camera import SimCamera
        from .telescope import SimTelescope

        Object.__init__(self, **kwargs)

        # get start time
        if time is None:
            time = Time.now()
        elif isinstance(time, str):
            time = Time(time)

        # calculate time offset
        self.time_offset = time - Time.now()

        # get telescope
        if telescope is None:
            self.telescope = SimTelescope(world=self)
        elif isinstance(telescope, SimTelescope):
            self.telescope = telescope
        elif isinstance(telescope, dict):
            self.telescope = create_object(telescope, world=self)
        else:
            raise ValueError("Invalid telescope.")

        # get camera
        if camera is None:
            self.camera = SimCamera(world=self)
        elif isinstance(camera, SimCamera):
            self.camera = camera
        elif isinstance(camera, dict):
            self.camera = create_object(camera, world=self)
        else:
            raise ValueError("Invalid camera.")

    async def open(self) -> None:
        """Open module."""
        await Object.open(self)

        # open telescope
        if hasattr(self.telescope, "open"):
            await self.telescope.open()

        # open camera
        if hasattr(self.telescope, "open"):
            await self.camera.open()

    async def close(self) -> None:
        """Close module."""
        await Object.close(self)

        # close telescope
        if hasattr(self.telescope, "close"):
            await self.telescope.close()

        # close camera
        if hasattr(self.camera, "close"):
            await self.camera.close()

    @property
    def time(self) -> Time:
        """Returns current time in simulation."""
        return Time.now() + self.time_offset

    @property
    def sun_alt(self) -> float:
        """Returns current solar altitude."""
        if self.observer is None:
            raise ValueError("No observer given.")
        return float(self.observer.sun_altaz(self.time).alt.degree)


__all__ = ["SimWorld"]



# import logging
# import threading
# from enum import Enum
# from typing import Union, List
# import pandas as pd
# import numpy as np
#
# from pyobs.modules import Module
# from pyobs.interfaces import ICamera, ISettings, IWindow, IBinning
# from pyobs.modules import timeout
# from pyobs.events import NewImageEvent, ExposureStatusChangedEvent
# from pyobs.utils.enums import ImageType
# from pyobs.utils.images import Image
# from pyobs.utils.photometry import SepPhotometry
#
# log = logging.getLogger(__name__)
#
#
# class AdaptiveCameraMode(Enum):
#     # find brightest star within radius around centre of image
#     CENTRE = 'centre',
#     # find brightest star in whole image
#     BRIGHTEST = 'brightest'
#
#
# class AdaptiveCamera(Module, ICamera, IWindow, IBinning, ISettings):
#     """A virtual camera for adaptive exposure times."""
#     # TODO: adapt this to new ICamera interface or remove!
#
#     def __init__(self, camera: str, mode: Union[str, AdaptiveCameraMode] = AdaptiveCameraMode.CENTRE, radius: int = 20,
#                  target_counts: int = 30000, min_exptime: float = 0.5, max_exptime: float = 60, history: int = 10,
#                  **kwargs: Any):
#         """Creates a new adaptive exposure time camera.
#
#         Args:
#             camera: Actual camera to use.
#             mode: Which mode to use to find star.
#             radius: Radius in px around centre for CENTRE mode.
#             target_counts: Counts to aim for in target.
#             min_exptime: Minimum exposure time.
#             max_exptime: Maximum exposure time.
#             history: Length of history.
#         """
#         Module.__init__(self, **kwargs)
#
#         # store
#         self._camera_name = camera
#         self._camera = None
#         self._mode = mode if isinstance(mode, AdaptiveCameraMode) else AdaptiveCameraMode(mode)
#         self._radius = radius
#         self._history: List[float] = []
#         self._max_history = history
#
#         # abort
#         self._abort = threading.Event()
#
#         # exposures
#         self._exp_time = None
#         self._exposure_count = None
#         self._exposures_done = None
#
#         # options
#         self._counts = target_counts
#         self._min_exp_time = min_exptime
#         self._max_exp_time = max_exptime
#
#         # SEP
#         self._sep = SepPhotometry()
#
#         # add thread
#         self._process_filename = None
#         self._process_lock = threading.RLock()
#         self._add_thread_func(self._process_thread, True)
#
#     def open(self):
#         """Open module."""
#         Module.open(self)
#
#         # subscribe to events
#         if self.comm:
#             self.comm.register_event(NewImageEvent)
#             self.comm.register_event(ExposureStatusChangedEvent, self._status_changed)
#
#         # get link to camera
#         self._camera = self.proxy(self._camera_name, ICamera)
#
#     @timeout('(exposure_time+10)*count')
#     def expose(self, exposure_time: int, image_type: ImageType, broadcast: bool = True, **kwargs: Any) -> str:
#         """Starts exposure and returns reference to image.
#
#         Args:
#             exposure_time: Exposure time in seconds.
#             image_type: Type of image.
#             broadcast: Broadcast existence of image.
#
#         Returns:
#             Name of image that was taken.
#         """
#
#         # reset
#         self._abort = threading.Event()
#         self._exp_time = exposure_time
#         self._exposures_done = 0
#
#         # loop exposures
#         return_filenames = []
#         self._history = [exposure_time]
#         count = 1
#         for i in range(count):
#             # abort?
#             if self._abort.is_set():
#                 break
#
#             # do exposure(s), never broadcast
#             log.info('Starting exposure with %d/%d for %.2fs...', i+1, count, self._exp_time)
#             filename = self._camera.expose(self._exp_time, image_type, 1, broadcast=False).wait()
#             self._exposures_done += 1
#
#             # store filename
#             return_filenames.append(filename)
#             with self._process_lock:
#                 if self._process_filename is None:
#                     self._process_filename = filename
#
#             # broadcast image path
#             if broadcast and self.comm:
#                 self.comm.send_event(NewImageEvent(filename, image_type))
#
#         # finished
#         self._exposure_count = None
#         self._exposures_done = None
#
#         # return filenames
#         return return_filenames
#
#     def abort(self, **kwargs: Any):
#         """Aborts the current exposure and sequence.
#
#         Returns:
#             Success or not.
#         """
#         self._abort.set()
#         self._camera.abort().wait()
#
#     def get_exposure_status(self, **kwargs: Any) -> ExposureStatus:
#         """Returns the current status of the camera, which is one of 'idle', 'exposing', or 'readout'.
#
#         Returns:
#             Current status of camera.
#         """
#         return self._camera.get_exposure_status().wait()
#
#     def abort_sequence(self, **kwargs: Any):
#         """Aborts the current sequence after current exposure.
#
#         Raises:
#             ValueError: If sequemce could not be aborted.
#         """
#         self._exposure_count = None
#         self._exposures_done = None
#         return self._camera.abort_sequence().wait()
#
#     def get_exposures_left(self, **kwargs: Any) -> int:
#         """Returns the remaining exposures.
#
#         Returns:
#             Remaining exposures
#         """
#         if self._exposures_done is None or self._exposure_count is None:
#             return 0
#         else:
#             return self._exposure_count - self._exposures_done
#
#     def get_exposure_time_left(self, **kwargs: Any) -> float:
#         """Returns the remaining exposure time on the current exposure in ms.
#
#         Returns:
#             Remaining exposure time in ms.
#         """
#         return self._camera.get_exposures_left().wait()
#
#     def get_exposure_progress(self, **kwargs: Any) -> float:
#         """Returns the progress of the current exposure in percent.
#
#         Returns:
#             Progress of the current exposure in percent.
#         """
#         return self._camera.get_exposure_progress().wait()
#
#     def _status_changed(self, event: ExposureStatusChangedEvent, sender: str, **kwargs: Any):
#         """Processing status change of camera.
#
#         Args:
#             event: Status change event.
#             sender: Name of sender.
#         """
#
#         # check sender
#         if sender == self._camera_name:
#             # resend event
#             ev = ExposureStatusChangedEvent(last=event.last, current=event.current)
#             self.comm.send_event(ev)
#
#     def _process_thread(self):
#         """Thread for processing images."""
#
#         # run until closing
#         while not self.closing.is_set():
#             # do we have an image?
#             with self._process_lock:
#                 filename = self._process_filename
#
#             # got something?
#             if filename is not None:
#                 # download image
#                 image = self.vfs.read_image(filename)
#
#                 # process it
#                 self._process_image(image)
#
#                 # reset image
#                 with self._process_lock:
#                     self._process_filename = None
#
#             # sleep a little
#             self.closing.wait(1)
#
#     def _process_image(self, image: Image):
#         """Process image.
#
#         Args:
#             image: Image to process.
#         """
#
#         # find peak count
#         peak = self._find_target(image)
#
#         # get exposure time from image in secs
#         exp_time = image.header['EXPTIME']
#
#         # scale exposure time
#         exp_time = int(exp_time * self._counts / peak)
#
#         # cut to limits
#         exp_time = max(min(exp_time, self._max_exp_time), self._min_exp_time)
#
#         # fill history
#         self._history.append(exp_time)
#         if len(self._history) > self._max_history:
#             self._history = self._history[-self._max_history:]
#
#         # set it
#         self._exp_time = int(np.mean(self._history))
#         log.info('Setting exposure time to %.3fs.', self._exp_time)
#
#     def _find_target(self, image: Image) -> int:
#         """Find target in image and return it's peak count.
#
#         Args:
#             image: Image to analyse.
#
#         Returns:
#             Peak count of target.
#         """
#
#         # find sources
#         sources: pd.DataFrame = self._sep.find_stars(image).to_pandas()
#
#         # which mode?
#         if self._mode == AdaptiveCameraMode.BRIGHTEST:
#             # sort by peak brightness and get first
#             sources.sort_values('peak', ascending=False, inplace=True)
#             row = sources.iloc[0]
#             log.info('Found brightest star at x=%.1f, y=%.1f with peak count of %d.', row['x'], row['y'], row['peak'])
#             return row['peak']
#
#         elif self._mode == AdaptiveCameraMode.CENTRE:
#             # get image centre
#             cx = image.header['CRPIX1'] if 'CRPIX1' in image.header else image.header['NAXIS1'] // 2
#             cy = image.header['CRPIX2'] if 'CRPIX2' in image.header else image.header['NAXIS2'] // 2
#
#             # filter all sources within radius around centre
#             r = self._radius
#             filtered = sources[(sources['x'] >= cx - r) & (sources['x'] <= cx + r) &
#                                (sources['y'] >= cy - r) & (sources['y'] <= cy + r)]
#
#             # sort by peak brightness and get first
#             filtered.sort_values('peak', ascending=False, inplace=True)
#             row = filtered.iloc[0]
#             log.info('Found brightest star at x=%.1f (dx=%.1f), y=%.1f (dy=%.1f) with peak count of %d.',
#                      row['x'], row['x'] - cx, row['y'], row['y'] - cy, row['peak'])
#             return row['peak']
#
#         else:
#             raise ValueError('Unknown target mode.')
#
#     def get_settings(self, **kwargs: Any) -> dict:
#         """Returns a dict of name->type pairs for settings."""
#         return {
#             'target_counts': 'int',
#             'min_exp_time': 'int',
#             'max_exp_time': 'int'
#         }
#
#     def get_setting_value(self, setting: str, **kwargs: Any):
#         """Returns the value of the given setting.
#
#         Args:
#             setting: Name of setting
#
#         Returns:
#             Current value
#
#         Raises:
#             KeyError if setting does not exist
#         """
#         if setting == 'target_counts':
#             return self._counts
#         elif setting == 'min_exp_time':
#             return self._min_exp_time
#         elif setting == 'max_exp_time':
#             return self._max_exp_time
#         else:
#             raise KeyError
#
#     def set_setting_value(self, setting: str, value, **kwargs: Any):
#         """Sets the value of the given setting.
#
#         Args:
#             setting: Name of setting
#             value: New value
#
#         Raises:
#             KeyError if setting does not exist
#         """
#         if setting == 'target_counts':
#             self._counts = value
#         elif setting == 'min_exp_time':
#             self._min_exp_time = value
#         elif setting == 'max_exp_time':
#             self._max_exp_time = value
#         else:
#             raise KeyError
#
#     def get_full_frame(self, **kwargs: Any) -> (int, int, int, int):
#         """Returns full size of CCD.
#
#         Returns:
#             Tuple with left, top, width, and height set.
#         """
#
#         # only do this, if wrapped camera doesn't support this
#         if isinstance(self._camera, IWindow):
#             return self._camera.get_full_frame().wait()
#         else:
#             return 0, 0, 100, 100
#
#     def set_window(self, left: int, top: int, width: int, height: int, **kwargs: Any):
#         """Set the camera window.
#
#         Args:
#             left: X offset of window.
#             top: Y offset of window.
#             width: Width of window.
#             height: Height of window.
#
#         Raises:
#             ValueError: If window could not be set.
#         """
#
#         # only do this, if wrapped camera doesn't support this
#         if isinstance(self._camera, IWindow):
#             self._camera.set_window(left, top, width, height).wait()
#
#     def get_window(self, **kwargs: Any) -> (int, int, int, int):
#         """Returns the camera window.
#
#         Returns:
#             Tuple with left, top, width, and height set.
#         """
#
#         # only do this, if wrapped camera doesn't support this
#         if isinstance(self._camera, IWindow):
#             return self._camera.get_window().wait()
#         else:
#             return 0, 0, 100, 100
#
#     def set_binning(self, x: int, y: int, **kwargs: Any):
#         """Set the camera binning.
#
#         Args:
#             x: X binning.
#             y: Y binning.
#
#         Raises:
#             ValueError: If binning could not be set.
#         """
#
#         # only do this, if wrapped camera doesn't support this
#         if isinstance(self._camera, IBinning):
#             self._camera.set_binning(x, y).wait()
#
#     def get_binning(self, **kwargs: Any) -> (int, int):
#         """Returns the camera binning.
#
#         Returns:
#             Tuple with x and y.
#         """
#
#         # only do this, if wrapped camera doesn't support this
#         if isinstance(self._camera, IBinning):
#             return self._camera.get_binning().wait()
#         else:
#             return 1, 1
#
#
# __all__ = ['AdaptiveCamera', 'AdaptiveCameraMode']



"""
Offsets
-------
"""

from .offsets import Offsets
from .astrometry import AstrometryOffsets
from .brighteststar import BrightestStarOffsets
from .nstar import NStarOffsets
from .projected import ProjectedOffsets
from .fitsheader import FitsHeaderOffsets


__all__ = [
    "Offsets",
    "AstrometryOffsets",
    "NStarOffsets",
    "ProjectedOffsets",
    "FitsHeaderOffsets",
    "BrightestStarOffsets",
]



from typing import Union, Dict, Any
import logging
import numpy as np
from astropy.io import fits
from numpy.typing import NDArray

from pyobs.images.processor import ImageProcessor
from pyobs.images import Image


log = logging.getLogger(__name__)


class AddMask(ImageProcessor):
    """Add mask to image."""

    __module__ = "pyobs.images.processors.misc"

    def __init__(self, masks: Dict[str, Dict[str, Union[NDArray[Any], str]]], **kwargs: Any):
        """Init an image processor that adds a mask to an image.

        Args:
            masks: Dictionary containing instrument->binning->mask, with binning as string, e.g. '1x1'.
        """
        ImageProcessor.__init__(self, **kwargs)

        # masks
        self._masks: Dict[str, Dict[str, NDArray[Any]]] = {}
        for instrument, group in masks.items():
            self._masks[instrument] = {}
            for binning, mask in group.items():
                if isinstance(mask, np.ndarray):
                    self._masks[instrument][binning] = mask
                elif isinstance(mask, str):
                    self._masks[instrument][binning] = fits.getdata(mask)
                else:
                    raise ValueError("Unknown mask format.")

    async def __call__(self, image: Image) -> Image:
        """Add mask to image.

        Args:
            image: Image to add mask to.

        Returns:
            Image with mask
        """

        # copy image
        img = image.copy()

        # add mask
        instrument = image.header["INSTRUME"]
        binning = "%dx%d" % (image.header["XBINNING"], image.header["YBINNING"])
        if binning in self._masks:
            img.mask = self._masks[instrument][binning].copy()
        else:
            log.warning("No mask found for binning of frame.")

        # finished
        return img


__all__ = ["AddMask"]



from typing import Tuple

from pyobs.images import Image


class FocusSeries:
    """Base class for focus series helper classes."""

    __module__ = "pyobs.utils.focusseries"

    def reset(self) -> None:
        """Reset focus series."""
        raise NotImplementedError

    def analyse_image(self, image: Image, focus_value: float) -> None:
        """Analyse given image.

        Args:
            image: Image to analyse
            focus_value: Value to fit along, e.g. focus value or its offset
        """
        raise NotImplementedError

    def fit_focus(self) -> Tuple[float, float]:
        """Fit focus from analysed images

        Returns:
            Tuple of new focus and its error
        """
        raise NotImplementedError


__all__ = ["FocusSeries"]



import argparse
import os
from typing import Type, Any

from pyobs import version


def init_cli():
    # init argument parsing
    # for all command line parameters we set the default to an environment variable,
    # so they can also be specified that way
    parser = argparse.ArgumentParser()

    # config
    parser.add_argument("config", type=str, help="Configuration file")

    # logging
    parser.add_argument(
        "--log-level", type=str, choices=["critical", "error", "warning", "info", "debug"], default="info"
    )
    parser.add_argument("-l", "--log-file", type=str, help="file to write log into")

    # debug stuff
    parser.add_argument("--debug-time", type=str, help="Fake time at start for pyobs to use")

    # version
    parser.add_argument("-v", "--version", action="version", version=version())

    # return it
    return parser


def parse_cli(parser: argparse.ArgumentParser):
    from pyobs.utils.time import Time

    # parse args
    args = parser.parse_args()

    # set debug time now
    if args.debug_time is not None:
        # calculate difference between now and given time
        delta = Time(args.debug_time) - Time.now()
        Time.set_offset_to_now(delta)

    # get full path of config
    if args.config:
        args.config = os.path.abspath(args.config)

    # finished
    return vars(args)


def start_daemon(app_class, pid_file=None, **kwargs: Any) -> None:
    """Start process as a daemon.

    Args:
        app_class: Class to create app from
        pid_file: Name of PID file.
    """
    import daemon
    from daemon import pidfile

    # get run directory
    run_dir = os.path.dirname(pid_file)

    # This launches the daemon in its context
    with daemon.DaemonContext(
        working_directory=run_dir, umask=0o002, pidfile=pidfile.TimeoutPIDLockFile(pid_file)
    ) as context:
        run(app_class, **kwargs)


def run(app_class: Type, **kwargs: Any) -> None:
    """Run a pyobs application with the given options.

    Args:
        app_class: Class to create app from
    """

    # create app and run it
    app = app_class(**kwargs)
    app.run()


def main() -> None:
    from pyobs.application import Application

    # init argument parsing and add PID/Daemon stuff
    parser = init_cli()
    parser.add_argument("-p", "--pid-file", type=str)

    # parse it
    args = parse_cli(parser)

    # run app
    if args["pid_file"]:
        start_daemon(app_class=Application, **args)
    else:
        run(app_class=Application, **args)


if __name__ == "__main__":
    main()



from __future__ import annotations
import asyncio
import inspect
import logging
from collections import Coroutine
from typing import Any, Union, Type, Dict, TYPE_CHECKING, Optional, Callable, TypeVar, overload, List

import pyobs.interfaces
from pyobs.events import Event, LogEvent, ModuleClosedEvent
from pyobs.interfaces import Interface
from pyobs.utils.parallel import Future
from .proxy import Proxy
from .commlogging import CommLoggingHandler

if TYPE_CHECKING:
    from pyobs.modules import Module

log = logging.getLogger(__name__)


ProxyType = TypeVar("ProxyType")


class Comm:
    """Base class for all Comm modules in pyobs."""

    __module__ = "pyobs.comm"

    def __init__(self, cache_proxies: bool = True):
        """Creates a comm module."""
        from pyobs.modules import Module

        self._proxies: Dict[str, Proxy] = {}
        self._module: Optional[Module] = None
        self._log_queue: asyncio.Queue[LogEvent] = asyncio.Queue()
        self._cache_proxies = cache_proxies
        self._logging_task: Optional[asyncio.Task] = None

        # add handler to global logger
        handler = CommLoggingHandler(self)
        handler.setLevel(logging.INFO)
        logging.getLogger().addHandler(handler)

    @property
    def module(self) -> Optional[Module]:
        """The module that this Comm object is attached to."""
        return self._module

    @module.setter
    def module(self, module: Module) -> None:
        """The module that this Comm object is attached to."""
        # if we have a _set_module method, call it
        self._set_module(module)

        # store module
        self._module = module

    def _set_module(self, module: Module) -> None:
        ...

    async def open(self) -> None:
        """Open module."""

        # start logging thread
        self._logging_task = asyncio.create_task(self._logging())

        # some events
        await self.register_event(ModuleClosedEvent, self._client_disconnected)

    async def close(self) -> None:
        """Close module."""

        # close thread
        if self._logging_task:
            self._logging_task.cancel()
        self._logging_task = None

    def _get_full_client_name(self, name: str) -> str:
        """Returns full name for given client.

        Some Comm modules may use short names for their clients. This methods returns the full name
        for a given short name.

        Args:
            name: Short name to get full name for.

        Returns:
            Full name for given client.
        """

        # this base class doesn't have short names
        return name

    async def _get_client(self, client: str) -> Optional[Union[Module, Proxy]]:
        """Get a proxy to the given client.

        Args:
            client: Name of client.

        Returns:
            Proxy class for given client.
        """

        # return module, if "main" is requested
        if client == "main":
            return self.module
        if client is None:
            return None

        # if client doesn't exist or we disabled caching, fetch a new proxy
        if client not in self._proxies or not self._cache_proxies:
            # get interfaces
            try:
                interfaces = await self.get_interfaces(client)
            except IndexError:
                return None

            # create new proxy
            proxy = Proxy(self, client, interfaces)
            self._proxies[client] = proxy

        # return proxy
        return self._proxies[client]

    @overload
    async def proxy(self, name_or_object: Union[str, object], obj_type: Type[ProxyType]) -> ProxyType:
        ...

    @overload
    async def proxy(self, name_or_object: Union[str, object], obj_type: Optional[Type[ProxyType]] = None) -> Any:
        ...

    async def proxy(
        self, name_or_object: Union[str, object], obj_type: Optional[Type[ProxyType]] = None
    ) -> Union[Any, ProxyType]:
        """Returns object directly if it is of given type. Otherwise get proxy of client with given name and check type.

        If name_or_object is an object:
            - If it is of type (or derived), return object.
            - Otherwise raise exception.
        If name_name_or_object is string:
            - Create proxy from name and raise exception, if it doesn't exist.
            - Check type and raise exception if wrong.
            - Return object.

        Args:
            name_or_object: Name of object or object itself.
            obj_type: Expected class of object.

        Returns:
            Object or proxy to object.

        Raises:
            ValueError: If proxy does not exist or wrong type.
        """

        if obj_type is not None and isinstance(name_or_object, obj_type):
            # return directly
            return name_or_object

        elif isinstance(name_or_object, str):
            # get proxy
            try:
                proxy = await self._get_client(name_or_object)
            except KeyError:
                raise ValueError(f"Could not get proxy for {name_or_object}.")

            # check it
            if proxy is None:
                raise ValueError('Could not create proxy for given name "%s".' % name_or_object)
            elif obj_type is None or isinstance(proxy, obj_type):
                return proxy
            else:
                raise ValueError(
                    'Proxy obtained from given name "%s" is not of requested type "%s".' % (name_or_object, obj_type)
                )

        else:
            # completely wrong...
            raise ValueError('Given parameter is neither a name nor an object of requested type "%s".' % obj_type)

    async def safe_proxy(
        self, name_or_object: Union[str, object], obj_type: Optional[Type[ProxyType]] = None
    ) -> Optional[Union[Any, ProxyType]]:
        """Calls proxy() in a safe way and returns None instead of raising an exception."""

        try:
            return await self.proxy(name_or_object, obj_type)
        except ValueError:
            return None

    async def _client_disconnected(self, event: Event, sender: str) -> bool:
        """Called when a client disconnects.

        Args:
            event: Disconnect event.
            sender: Name of client that disconnected.

        """

        # if a client disconnects, we remove its proxy
        if sender in self._proxies:
            del self._proxies[sender]
        return True

    @property
    def name(self) -> Optional[str]:
        """Name of this client."""
        raise NotImplementedError

    @property
    def clients(self) -> List[str]:
        """Returns list of currently connected clients.

        Returns:
            (list) List of currently connected clients.
        """
        raise NotImplementedError

    async def clients_with_interface(self, interface: Type[Interface]) -> List[str]:
        """Returns list of currently connected clients that implement the given interface.

        Args:
            interface: Interface to search for.

        Returns:
            (list) List of currently connected clients that implement the given interface.
        """
        return [c for c in self.clients if await self._supports_interface(c, interface)]

    async def get_interfaces(self, client: str) -> List[Type[Interface]]:
        """Returns list of interfaces for given client.

        Args:
            client: Name of client.

        Returns:
            List of supported interfaces.

        Raises:
            IndexError: If client cannot be found.
        """
        raise NotImplementedError

    async def _supports_interface(self, client: str, interface: Type[Interface]) -> bool:
        """Checks, whether the given client supports the given interface.

        Args:
            client: Client to check.
            interface: Interface to check.

        Returns:
            Whether or not interface is supported.
        """
        raise NotImplementedError

    @staticmethod
    def _interface_names_to_classes(interfaces: List[str]) -> List[Type[Interface]]:
        """Converts a list of interface names to interface classes.

        Args:
            interfaces: List of interface names.

        Returns:
            List of interface classes.
        """

        # get interface classes
        inspection = inspect.getmembers(pyobs.interfaces, predicate=inspect.isclass)

        # loop interfaces
        interface_classes = []
        for interface_name in interfaces:
            # loop all classes
            found = False
            for cls_name, cls in inspection:
                # class needs to face same name and implement Interface
                if interface_name == cls_name and issubclass(cls, Interface):
                    # found it!
                    found = True

                    # then add it to the list of all interfaces
                    interface_classes.append(cls)

                    # there can only be one...
                    break

            # not found?
            if not found:
                log.error('Could not find interface "%s" for client.', interface_name)
        return interface_classes

    async def execute(self, client: str, method: str, signature: inspect.Signature, *args: Any) -> Future:
        """Execute a given method on a remote client.

        Args:
            client (str): ID of client.
            method (str): Method to call.
            signature: Method signature.
            *args: List of parameters for given method.

        Returns:
            Passes through return from method call.
        """
        raise NotImplementedError

    async def _logging(self) -> None:
        """Background thread for handling the logging."""
        # run until closing
        while True:
            # get item (maybe wait for it) and send it
            try:
                entry = await self._log_queue.get()
                await self.send_event(entry)
            except asyncio.CancelledError:
                break
            except:
                log.exception("bla")
                pass

    def log_message(self, entry: LogEvent) -> None:
        """Send a log message to other clients.

        Args:
            entry (LogEvent): Log event to send.
        """
        self._log_queue.put_nowait(entry)

    async def send_event(self, event: Event) -> None:
        """Send an event to other clients.

        Args:
            event (Event): Event to send
        """
        pass

    async def register_event(
        self, event_class: Type[Event], handler: Optional[Callable[[Event, str], Coroutine[Any, Any, bool]]] = None
    ) -> None:
        """Register an event type. If a handler is given, we also receive those events, otherwise we just
        send them.

        Args:
            event_class: Class of event to register.
            handler: Event handler method.
        """
        pass


__all__ = ["Comm"]



import io
import logging
from typing import Dict, Any, AnyStr

from .file import VFSFile


log = logging.getLogger(__name__)


class MemoryFile(VFSFile):
    """A file stored in memory."""

    __module__ = "pyobs.vfs"

    """Global buffer."""
    _buffer: Dict[str, AnyStr] = {}

    def __init__(self, name: str, mode: str = "r", **kwargs: Any):
        """Open/create a file in memory.

        Args:
            name: Name of file.
            mode: Open mode.
        """

        # init
        io.RawIOBase.__init__(self)

        # store
        self._filename = name
        self._mode = mode
        self._pos = 0
        self._open = True

        # overwrite?
        if "w" in mode:
            MemoryFile._buffer[name] = b"" if "b" in mode else ""

    async def read(self, n: int = -1) -> AnyStr:
        """Read number of bytes from stream.

        Args:
            n: Number of bytes to read, -1  reads until end of data.

        Returns:
            Data read from stream.
        """

        # check size
        if n == -1:
            data = MemoryFile._buffer[self._filename]
            self._pos = len(data) - 1
        else:
            # extract data to read
            data = MemoryFile._buffer[self._filename][self._pos : self._pos + n]
            self._pos += n

        # return data
        return data

    async def write(self, s: AnyStr) -> None:
        """Write data into the stream.

        Args:
            s: Bytes of data to write.
        """
        MemoryFile._buffer[self._filename] += s

    async def close(self) -> None:
        """Close stream."""

        # set flag
        self._open = False

    @property
    def closed(self) -> bool:
        """Whether stream is closed."""
        return not self._open


__all__ = ["MemoryFile"]



from __future__ import annotations
from typing import Any, TYPE_CHECKING, Optional, Dict

from pyobs.object import Object
from .task import Task

if TYPE_CHECKING:
    from .taskschedule import TaskSchedule
    from .taskarchive import TaskArchive


class TaskRunner(Object):
    def __init__(
        self,
        scripts: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ):
        """Creates a new LCO scheduler.

        Args:
            scripts: External scripts
        """
        Object.__init__(self, **kwargs)

        # store stuff
        self.scripts = scripts

    async def can_run(self, task: Task) -> bool:
        """Checks, whether this task could run now.

        Returns:
            True, if task can run now.
        """
        return await task.can_run(self.scripts)

    async def run_task(
        self, task: Task, task_schedule: Optional[TaskSchedule] = None, task_archive: Optional[TaskArchive] = None
    ) -> bool:
        """Run a task.

        Args:
            task: Task to run
            task_schedule: Schedule.
            task_archive: Archive.

        Returns:
            Success or not
        """

        # run task
        await task.run(task_runner=self, task_schedule=task_schedule, task_archive=task_archive, scripts=self.scripts)

        # finish
        return True


__all__ = ["TaskRunner"]



import logging
from typing import List, Any

from .filelist import FileList


log = logging.getLogger(__name__)


class TestingFileList(FileList):
    """File list for testing."""

    __module__ = "pyobs.vfs.filelists"

    def __init__(self, *args: Any, **kwargs: Any):
        pass

    def __call__(self, filename: str) -> List[str]:
        return [__file__]


__all__ = ["TestingFileList"]



from abc import ABCMeta

from .IRoof import IRoof
from .IPointingAltAz import IPointingAltAz


class IDome(IRoof, IPointingAltAz, metaclass=ABCMeta):
    """The module controls a dome, i.e. a :class:`~pyobs.interfaces.IRoof` with a rotating roof."""

    __module__ = "pyobs.interfaces"
    pass


__all__ = ["IDome"]



import asyncio
import fnmatch
import os
from pathlib import PurePosixPath
from typing import Any, Optional, Iterator, BinaryIO, IO, AnyStr, cast, List

from .file import VFSFile


class LocalFile(VFSFile):
    """Wraps a local file with the virtual file system."""

    __module__ = "pyobs.vfs"

    def __init__(self, name: str, mode: str = "r", root: Optional[str] = None, mkdir: bool = True, **kwargs: Any):
        """Open a local file.

        Args:
            name: Name of file.
            mode: Open mode.
            root: Root to prefix name with for absolute path in filesystem.
            mkdir: Whether or not to create non-existing paths automatically.
        """

        # no root given?
        if root is None:
            raise ValueError("No root directory given.")

        # filename is not allowed to start with a / or contain ..
        if name.startswith("/") or ".." in name:
            raise ValueError("Only files within root directory are allowed.")

        # build filename
        self.filename = name
        full_path = os.path.join(root, name)

        # need to create directory?
        path = os.path.dirname(full_path)
        if not os.path.exists(path):
            if mkdir:
                os.makedirs(path)
            else:
                raise ValueError("Cannot write into sub-directory with disabled mkdir option.")

        # file object
        self.fd: IO[Any] = open(full_path, mode)

    async def close(self) -> None:
        if self.fd:
            self.fd.close()

    async def read(self, n: int = -1) -> AnyStr:
        if self.fd is None:
            raise OSError
        return cast(AnyStr, self.fd.read(n))

    async def write(self, s: AnyStr) -> None:
        if self.fd is None:
            raise OSError
        self.fd.write(s)

    @staticmethod
    async def local_path(path: str, **kwargs: Any) -> str:
        """Returns local path of given path.

        Args:
            path: Path to list.
            kwargs: Parameters for specific file implementation (same as __init__).

        Returns:
            Local path.
        """

        # get settings
        root = kwargs["root"]

        # return path
        return os.path.join(root, path)

    @staticmethod
    async def listdir(path: str, **kwargs: Any) -> List[str]:
        """Returns content of given path.

        Args:
            path: Path to list.
            kwargs: Parameters for specific file implementation (same as __init__).

        Returns:
            List of files in path.
        """

        # get settings
        root = kwargs["root"]

        # get path and return list
        full_path = PurePosixPath(root) / path
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, os.listdir, str(full_path))

    @staticmethod
    async def find(path: str, pattern: str, **kwargs: Any) -> List[str]:
        """Find files by pattern matching.

        Args:
            path: Path to search in.
            pattern: Pattern to search for.

        Returns:
            List of found files.
        """

        # get root from kwargs
        if "root" not in kwargs:
            raise ValueError("No root directory given.")
        root = kwargs["root"]

        # build full path
        full_path = os.path.join(root, path)

        # loop directories
        files = []
        for cur, dirnames, filenames in os.walk(full_path):
            for filename in fnmatch.filter(filenames, pattern):
                files += [os.path.relpath(os.path.join(cur, filename), full_path)]
        return files

    @staticmethod
    async def remove(path: str, *args: Any, **kwargs: Any) -> bool:
        """Remove file at given path.

        Args:
            path: Path of file to delete.

        Returns:
            Success or not.
        """

        # get root from kwargs
        root = kwargs["root"]

        # build full path and remove
        full_path = os.path.join(root, path)
        try:
            os.remove(full_path)
            return True
        except (FileNotFoundError, IsADirectoryError):
            return False

    @classmethod
    async def exists(cls, path: str, root: str = "", *args: Any, **kwargs: Any) -> bool:
        """Checks, whether a given path or file exists.

        Args:
            path: Path to check.
            root: VFS root.

        Returns:
            Whether it exists or not
        """

        # build full path
        full_path = os.path.join(root, path)

        # check
        return os.path.exists(full_path)


__all__ = ["LocalFile"]



from abc import ABCMeta, abstractmethod
from typing import Tuple, Any

from .interface import Interface


class IPointingHGS(Interface, metaclass=ABCMeta):
    """The module can move to Mu/Psi coordinates, usually combined with :class:`~pyobs.interfaces.ITelescope`."""

    __module__ = "pyobs.interfaces"

    @abstractmethod
    async def move_hgs_lon_lat(self, lon: float, lat: float, **kwargs: Any) -> None:
        """Moves on given coordinates.

        Args:
            lon: Longitude in deg to track.
            lat: Latitude in deg to track.

        Raises:
            MoveError: If device could not be moved.
        """
        ...

    @abstractmethod
    async def get_hgs_lon_lat(self, **kwargs: Any) -> Tuple[float, float]:
        """Returns current longitude and latitude position.

        Returns:
            Tuple of current lon, lat in degrees.
        """
        ...


__all__ = ["IPointingHGS"]



import asyncio
from functools import partial
from typing import Any
from astropy.stats import sigma_clipped_stats
import logging
import numpy as np
from photutils import CircularAnnulus, CircularAperture, aperture_photometry

from .photometry import Photometry
from pyobs.images import Image

log = logging.getLogger(__name__)


class PhotUtilsPhotometry(Photometry):
    """Perform photometry using PhotUtils."""

    __module__ = "pyobs.images.processors.photometry"

    def __init__(
        self,
        threshold: float = 1.5,
        minarea: int = 5,
        deblend_nthresh: int = 32,
        deblend_cont: float = 0.005,
        clean: bool = True,
        clean_param: float = 1.0,
        **kwargs: Any,
    ):
        """Initializes an aperture photometry based on PhotUtils.

        Args:
            threshold: Threshold pixel value for detection.
            minarea: Minimum number of pixels required for detection.
            deblend_nthresh: Number of thresholds used for object deblending.
            deblend_cont: Minimum contrast ratio used for object deblending.
            clean: Perform cleaning?
            clean_param: Cleaning parameter (see SExtractor manual).
            *args:
            **kwargs:
        """
        Photometry.__init__(self, **kwargs)

        # store
        self.threshold = threshold
        self.minarea = minarea
        self.deblend_nthresh = deblend_nthresh
        self.deblend_cont = deblend_cont
        self.clean = clean
        self.clean_param = clean_param

    async def __call__(self, image: Image) -> Image:
        """Do aperture photometry on given image.

        Args:
            image: Image to do aperture photometry on.

        Returns:
            Image with attached catalog.
        """
        loop = asyncio.get_running_loop()

        # no pixel scale given?
        if image.pixel_scale is None:
            log.warning("No pixel scale provided by image.")
            return image

        # fetch catalog
        if image.catalog is None:
            log.warning("No catalog in image.")
            return image
        sources = image.catalog.copy()

        # get positions
        positions = [(x - 1, y - 1) for x, y in sources.iterrows("x", "y")]

        # perform aperture photometry for diameters of 1" to 8"
        for diameter in [1, 2, 3, 4, 5, 6, 7, 8]:
            # extraction radius in pixels
            radius = diameter / 2.0 / image.pixel_scale
            if radius < 1:
                continue

            # defines apertures
            aperture = CircularAperture(positions, r=radius)
            annulus_aperture = CircularAnnulus(positions, r_in=2 * radius, r_out=3 * radius)
            annulus_masks = annulus_aperture.to_mask(method="center")

            # loop annuli
            bkg_median = []
            for m in annulus_masks:
                annulus_data = m.multiply(image.data)
                annulus_data_1d = annulus_data[m.data > 0]
                _, median_sigclip, _ = sigma_clipped_stats(annulus_data_1d)
                bkg_median.append(median_sigclip)

            # do photometry
            phot = await loop.run_in_executor(
                None, partial(aperture_photometry, image.data, aperture, mask=image.mask, error=image.uncertainty)
            )

            # calc flux
            bkg_median_np = np.array(bkg_median)
            aper_bkg = bkg_median_np * aperture.area
            sources["fluxaper%d" % diameter] = phot["aperture_sum"] - aper_bkg
            if "aperture_sum_err" in phot.columns:
                sources["fluxerr%d" % diameter] = phot["aperture_sum_err"]
            sources["bkgaper%d" % diameter] = bkg_median_np

        # copy image, set catalog and return it
        img = image.copy()
        img.catalog = sources
        return img


__all__ = ["PhotUtilsPhotometry"]



from abc import ABCMeta
from collections import Coroutine
from datetime import datetime
import io
import logging
import time
import asyncio
from typing import Dict, Any, Tuple, NamedTuple, Optional, List
import numpy as np
import PIL.Image
from aiohttp import web
from numpy.typing import NDArray

from pyobs.modules import Module, timeout
from pyobs.interfaces import IVideo, IImageType, IExposureTime
from pyobs.events import NewImageEvent
from pyobs.images import Image
from pyobs.mixins.fitsheader import ImageFitsHeaderMixin
from pyobs.utils.cache import DataCache
from pyobs.utils.enums import ImageType

log = logging.getLogger(__name__)

INDEX_HTML = """
<!DOCTYPE html>
<html lang="en">
  <head>
    <meta charset="UTF-8">
    <title>Title</title>
  </head>
  <body>
    <img src="/video.mjpg" width="100%">
  </body>
</html>

"""


async def calc_expose_timeout(webcam: IExposureTime, *args: Any, **kwargs: Any) -> float:
    """Calculates timeout for grabe_image()."""
    if hasattr(webcam, "get_exposure_time"):
        return 2.0 * await webcam.get_exposure_time() + 30.0
    else:
        return 30.0


class NextImage(NamedTuple):
    date_obs: str
    image_type: ImageType
    header_futures: Dict[str, Coroutine[Dict[str, Tuple[Any, str]], None, None]]
    broadcast: bool


class ImageRequest:
    def __init__(self, broadcast: bool = True):
        self.broadcast: bool = broadcast
        self.image: Optional[Image] = None
        self.filename: Optional[str] = None


class LastImage(NamedTuple):
    data: NDArray[Any]
    image: Optional[Image]
    jpeg: Optional[bytes]
    filename: Optional[str]


class BaseVideo(Module, ImageFitsHeaderMixin, IVideo, IImageType, metaclass=ABCMeta):
    """Base class for all webcam modules."""

    __module__ = "pyobs.modules.camera"

    def __init__(
        self,
        http_port: int = 37077,
        interval: float = 0.5,
        video_path: str = "/webcam/video.mjpg",
        filenames: str = "/webcam/pyobs-{DAY-OBS|date:}-{FRAMENUM|string:04d}.fits",
        fits_namespaces: Optional[List[str]] = None,
        fits_headers: Optional[Dict[str, Any]] = None,
        centre: Optional[Tuple[float, float]] = None,
        rotation: float = 0.0,
        cache_size: int = 5,
        live_view: bool = True,
        flip: bool = False,
        sleep_time: int = 600,
        **kwargs: Any,
    ):
        """Creates a new BaseWebcam.

        On the receiving end, a VFS root with a HTTPFile must exist with the same name as in image_path and video_path,
        i.e. "webcam" in the default settings.

        Args:
            http_port: HTTP port for webserver.
            exposure_time: Initial exposure time.
            interval: Min interval for grabbing images.
            video_path: VFS path to video.
            filename: Filename pattern for FITS images.
            fits_namespaces: List of namespaces for FITS headers that this camera should request.
            fits_headers: Additional FITS headers.
            centre: (x, y) tuple of camera centre.
            rotation: Rotation east of north.
            cache_size: Size of cache for previous images.
            live_view: If True, live view is served via web server.
            flip: Whether to flip around Y axis.
            sleep_time: Time in s with inactivity after which the camera should go to sleep.
        """
        Module.__init__(self, **kwargs)
        ImageFitsHeaderMixin.__init__(
            self,
            fits_namespaces=fits_namespaces,
            fits_headers=fits_headers,
            centre=centre,
            rotation=rotation,
            filenames=filenames,
        )

        # store
        self._is_listening = False
        self._port = http_port
        self._interval = interval
        self._new_image_event = asyncio.Event()
        self._video_path = video_path
        self._frame_num = 0
        self._live_view = live_view
        self._image_type = ImageType.OBJECT
        self._image_request_lock = asyncio.Lock()
        self._image_requests: List[ImageRequest] = []
        self._next_image: Optional[NextImage] = None
        self._last_image: Optional[LastImage] = None
        self._last_time = 0.0
        self._flip = flip
        self._sleep_time = sleep_time

        # active
        self._active = False
        self._active_time = 0.0
        self.add_background_task(self._active_update)

        # image cache
        self._cache = DataCache(cache_size)

        # define web server
        self._app = web.Application()
        self._app.add_routes(
            [
                web.get("/", self.web_handler),
                web.get("/video.mjpg", self.video_handler),
                web.get("/{filename}", self.image_handler),
            ]
        )
        self._runner = web.AppRunner(self._app)
        self._site: Optional[web.TCPSite] = None

    async def open(self) -> None:
        """Open module."""
        await Module.open(self)

        # start listening
        log.info("Starting HTTP file cache on port %d...", self._port)
        await self._runner.setup()
        self._site = web.TCPSite(self._runner, "0.0.0.0", self._port)
        await self._site.start()
        self._is_listening = True

    async def close(self) -> None:
        """Close server"""
        await Module.close(self)

        # stop server
        await self._runner.cleanup()

    @property
    def opened(self) -> bool:
        """Whether the server is started."""
        return self._is_listening

    async def web_handler(self, request: web.Request) -> web.Response:
        """Handles access to / and returns HTML page.

        Args:
            request: Request to respond to.

        Returns:
            Response containing web page.
        """
        return web.Response(text=INDEX_HTML, content_type="text/html")

    async def video_handler(self, request: web.Request) -> web.StreamResponse:
        """Handles access to /video.mjpg and returns the video.

        Args:
            request: Request to respond to.

        Returns:
            Response containing video stream.
        """

        # create response
        response = web.StreamResponse()
        response.content_type = "multipart/x-mixed-replace; boundary=--jpgboundary"
        await response.prepare(request)

        last_num = None
        last_time = 0.0
        interval = 1.0
        while True:
            # not reached interval?
            if time.time() < last_time + interval:
                await asyncio.sleep(0.01)
                continue

            # get image
            num, image = await self.image_jpeg()
            if image is None:
                await asyncio.sleep(0.01)
                continue

            # is it actually a new image?
            if num == last_num:
                await asyncio.sleep(0.01)
                continue

            # now send image!
            last_num = num
            last_time = time.time()
            await response.write(b"--jpgboundary\r\nContent-type: image/jpeg\r\n\r\n" + image + b"\r\n")

        # return response
        return response

    async def image_handler(self, request: web.Request) -> web.Response:
        """Handles access to /* and returns a specified image.

        Args:
            request: Request to respond to.

        Returns:
            Response containing image.
        """

        # get filename
        filename = request.match_info["filename"]

        # get data
        if filename not in self._cache:
            raise web.HTTPNotFound()
        data = self._cache[filename]

        # send it
        log.info(f"Serving file {filename}.")
        return web.Response(body=data, content_type="image/fits")

    @property
    def camera_active(self) -> bool:
        """Whether camera is currently active."""
        return self._active

    async def activate_camera(self) -> None:
        """Activate camera."""
        self._active_time = time.time()
        if not self._active:
            await self._activate_camera()
        self._active = True

    async def deactivate_camera(self) -> None:
        """Deactivate camera."""
        self._active_time = 0
        if self._active:
            await self._deactivate_camera()
        self._active = False

    async def _activate_camera(self) -> None:
        """Can be overridden by derived class to implement inactivity sleep"""
        pass

    async def _deactivate_camera(self) -> None:
        """Can be overridden by derived class to implement inactivity sleep"""
        pass

    async def _active_update(self) -> None:
        """Checking active status regularly."""
        self._active_time = time.time()
        while True:
            # go to sleep?
            if time.time() - self._active_time > self._sleep_time and self._active:
                await self.deactivate_camera()

            # wait a little for next check
            await asyncio.sleep(1)

    async def image_jpeg(self) -> Tuple[Optional[int], Optional[bytes]]:
        """Return image as jpeg."""

        # activate camera, first image will most probably be None
        await self.activate_camera()

        # return what we got
        return self._frame_num, None if self._last_image is None else self._last_image.jpeg

    @staticmethod
    def create_jpeg(data: NDArray[Any]) -> bytes:
        """Create a JPEG ge from a numpy array and return as bytes.

        Args:
            data: Numpy array to convert to JPEG.

        Returns:
            Bytes containing JPEG image.
        """

        # uint16?
        if data.dtype == np.uint16:
            # TODO: find a better way to convert to uint8
            data = (data / 256).astype(np.uint8)

        # write to jpeg
        with io.BytesIO() as output:
            PIL.Image.fromarray(data).save(output, format="jpeg")
            return output.getvalue()

    async def _set_image(self, data: NDArray[Any]) -> None:
        """Create FITS and JPEG images from data."""

        # flip image?
        if self._flip:
            data: NDArray[Any] = np.flip(data, axis=0)  # type: ignore

        # got a requested image in the queue?
        image, filename = None, None
        if self._next_image is not None:
            # create image and reset
            image, filename = await self._create_image(data, self._next_image)
            self._next_image = None
            async with self._image_request_lock:
                for req in self._image_requests:
                    req.image = image
                    req.filename = filename

        # convert to jpeg only if we need live view
        now = time.time()
        jpeg = None
        if self._live_view:
            # check interval
            if now - self._last_time > self._interval:
                # write to buffer and reset interval
                loop = asyncio.get_running_loop()
                jpeg = await loop.run_in_executor(None, self.create_jpeg, data)
                self._last_time = now

        # store both
        self._last_image = LastImage(data=data, image=image, jpeg=jpeg, filename=filename)
        self._frame_num += 1

        # signal it
        self._new_image_event.set()
        self._new_image_event = asyncio.Event()

        # prepare next image
        if len(self._image_requests) > 0:
            # broadcast?
            broadcast = any([req.broadcast for req in self._image_requests])

            # store everything
            logging.info("Preparing to catch next image...")
            self._next_image = NextImage(
                date_obs=datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%S.%f"),
                image_type=self._image_type,
                header_futures=await self.request_fits_headers(),
                broadcast=broadcast,
            )

            # reset
            self._image_request = None

    async def _create_image(self, data: NDArray[Any], next_image: NextImage) -> Tuple[Image, str]:
        """Create an Image object from numpy array.

        Args:
            data: Numpy array to convert to Image.

        Returns:
            Tuple with image itself and the filename.
        """

        # create image
        flipped: NDArray[Any] = np.flip(data, axis=0)  # type: ignore
        image = Image(flipped)
        image.header["DATE-OBS"] = next_image.date_obs
        image.header["IMAGETYP"] = next_image.image_type.value

        # add fits headers and format filename
        await self.add_requested_fits_headers(image, next_image.header_futures)
        await self.add_fits_headers(image)

        # finish it up
        return await self._finish_image(image, next_image.broadcast, next_image.image_type)

    async def _finish_image(self, image: Image, broadcast: bool, image_type: ImageType) -> Tuple[Image, str]:
        """Finish up an image at the end of _create_image.

        Args:
            image: Image to finish up.
            broadcast: Whether to broadcast it.
            image_type: Type of image.

        Returns:
            Tuple with image itself and the filename.
        """

        # format filename
        filename = self.format_filename(image)
        if filename is None:
            filename = "image.fits"

        # store it and return filename
        log.info("Writing image %s to cache...", filename)
        self._cache[image.header["FNAME"]] = image.to_bytes()

        # broadcast image path
        if broadcast and self.comm:
            log.info("Broadcasting image ID...")
            await self.comm.send_event(NewImageEvent(filename, image_type))

        # finished
        return image, filename

    @timeout(calc_expose_timeout)
    async def grab_image(self, broadcast: bool = True, **kwargs: Any) -> str:
        """Grabs an image ans returns reference.

        Args:
            broadcast: Broadcast existence of image.

        Returns:
            Name of image that was taken.
        """

        # activate camera
        await self.activate_camera()

        # acquire lock
        async with self._image_request_lock:
            # request new image
            image_request = ImageRequest(broadcast)
            self._image_requests.append(image_request)

        # we want an image that starts exposing AFTER now, so we wait for the current image to finish.
        log.info("Waiting for image to finish...")
        while image_request.image is None:
            await asyncio.sleep(0.1)

        # remove from list
        self._image_requests.remove(image_request)

        # no image?
        if image_request.image is None or image_request.filename is None:
            raise ValueError("Could not take image.")

        # finished
        return image_request.filename

    async def get_video(self, **kwargs: Any) -> str:
        """Returns path to video.

        Returns:
            Path to video.
        """
        return self._video_path

    async def set_image_type(self, image_type: ImageType, **kwargs: Any) -> None:
        """Set the image type.

        Args:
            image_type: New image type.
        """
        log.info("Setting image type to %s...", image_type)
        self._image_type = image_type

    async def get_image_type(self, **kwargs: Any) -> ImageType:
        """Returns the current image type.

        Returns:
            Current image type.
        """
        return self._image_type


__all__ = ["BaseVideo", "NextImage"]



import asyncio
import datetime
import logging
from abc import ABCMeta, abstractmethod
from typing import Tuple, Optional, Dict, Any, NamedTuple, List
from astropy.io import fits

from pyobs.mixins.fitsheader import SpectrumFitsHeaderMixin
from pyobs.utils.enums import ExposureStatus
from pyobs.modules import Module
from pyobs.events import NewSpectrumEvent, ExposureStatusChangedEvent
from pyobs.interfaces import ISpectrograph
from pyobs.modules import timeout

log = logging.getLogger(__name__)


class ExposureInfo(NamedTuple):
    """Info about a running exposure."""

    start: datetime.datetime


class BaseSpectrograph(Module, SpectrumFitsHeaderMixin, ISpectrograph, metaclass=ABCMeta):
    """Base class for all spectrograph modules."""

    __module__ = "pyobs.modules.camera"

    def __init__(
        self,
        fits_headers: Optional[Dict[str, Any]] = None,
        filenames: str = "/cache/pyobs-{DAY-OBS|date:}-{FRAMENUM|string:04d}.fits.gz",
        fits_namespaces: Optional[List[str]] = None,
        **kwargs: Any,
    ):
        """Creates a new BaseCamera.

        Args:
            fits_headers: Additional FITS headers.
            flip: Whether or not to flip the image along its first axis.
            filenames: Template for file naming.
            fits_namespaces: List of namespaces for FITS headers that this camera should request
        """
        Module.__init__(self, **kwargs)
        SpectrumFitsHeaderMixin.__init__(
            self, fits_namespaces=fits_namespaces, fits_headers=fits_headers, filenames=filenames, **kwargs
        )

        # init camera
        self._exposure: Optional[ExposureInfo] = None
        self._spectrograph_status = ExposureStatus.IDLE

        # multi-threading
        self._expose_lock = asyncio.Lock()
        self.expose_abort = asyncio.Event()

        # check
        if self.comm is None:
            log.warning("No comm module given, will not be able to signal new images!")

    @abstractmethod
    async def _expose(self, abort_event: asyncio.Event) -> fits.HDUList:
        """Actually do the exposure, should be implemented by derived classes.

        Args:
            abort_event: Event that gets triggered when exposure should be aborted.

        Returns:
            The actual image and, if present, a filename.

        Raises:
            ValueError: If exposure was not successful.
        """
        ...

    async def __expose(self, broadcast: bool) -> Tuple[Optional[fits.HDUList], Optional[str]]:
        """Wrapper for a single exposure.

        Args:
            broadcast: Whether or not the new image should be broadcasted.

        Returns:
            Tuple of the image itself and its filename.

        Raises:
            ValueError: If exposure was not successful.
        """

        # request fits headers
        header_futures_before = await self.request_fits_headers(before=True)

        # do the exposure
        self._exposure = ExposureInfo(start=datetime.datetime.utcnow())
        try:
            hdulist = await self._expose(abort_event=self.expose_abort)
            if hdulist is None or hdulist[0].data is None:
                self._exposure = None
                return None, None
        except:
            # exposure was not successful (aborted?), so reset everything
            self._exposure = None
            raise

        # request fits headers again
        header_futures_after = await self.request_fits_headers(before=False)

        # add HDU name
        hdulist[0].header["EXTNAME"] = "SCI"

        # add fits headers and format filename
        await self.add_requested_fits_headers(hdulist[0], header_futures_before)
        await self.add_requested_fits_headers(hdulist[0], header_futures_after)
        await self.add_fits_headers(hdulist[0])

        # format filename
        filename = self.format_filename(hdulist[0])

        # don't want to save?
        if filename is None:
            return hdulist, None

        # store spectrum
        await self.store_spectrum(hdulist, filename, broadcast)

        # return spectrum and unique
        self._exposure = None
        log.info("Finished spectrum %s.", filename)
        return hdulist, filename

    async def store_spectrum(self, hdulist: fits.HDUList, filename: str, broadcast: bool) -> None:
        """Store spectrum at given destination.
        Can be overwritten by derived classes to custom store a file. In those cases, this version should be called!

        Args:
            hdulist: HDU list with spectrum.
            filename: Name to store file with.
            broadcast: Whether to broadcast new spectrum.
        """

        # upload file
        try:
            log.info("Uploading spectrum to file server...")
            await self.vfs.write_fits(filename, hdulist)
        except FileNotFoundError:
            raise ValueError("Could not upload spectrum.")

        # broadcast image path
        if broadcast and self.comm:
            log.info("Broadcasting spectrum ID...")
            await self.comm.send_event(NewSpectrumEvent(filename))

    @timeout(10)
    async def grab_spectrum(self, broadcast: bool = True, **kwargs: Any) -> str:
        """Grabs a spectrum and returns reference.

        Args:
            broadcast: Broadcast existence of image.

        Returns:
            Name of image that was taken.
        """

        # are we exposing?
        if self._spectrograph_status != ExposureStatus.IDLE:
            raise ValueError("Cannot start new exposure because spectrograph is not idle.")
        await self._change_exposure_status(ExposureStatus.EXPOSING)

        # expose
        hdu, filename = await self.__expose(broadcast)
        if hdu is None:
            raise ValueError("Could not take spectrum.")
        else:
            if filename is None:
                raise ValueError("Spectrum has not been saved, so cannot be retrieved by filename.")

        # return filename
        await self._change_exposure_status(ExposureStatus.IDLE)
        return filename

    async def _change_exposure_status(self, status: ExposureStatus) -> None:
        """Change exposure status and send event,

        Args:
            status: New exposure status.
        """

        # send event, if it changed
        if self._spectrograph_status != status:
            await self.comm.send_event(ExposureStatusChangedEvent(last=self._spectrograph_status, current=status))

        # set it
        self._spectrograph_status = status

    async def get_exposure_status(self, **kwargs: Any) -> ExposureStatus:
        """Returns the current status of the spectrograph, which is one of 'idle', 'exposing', or 'readout'.

        Returns:
            Current status of spectrograph.
        """
        return self._spectrograph_status

    async def abort(self, **kwargs: Any) -> None:
        """Aborts the current exposure.

        Raises:
            ValueError: If exposure could not be aborted.
        """
        pass


__all__ = ["BaseSpectrograph"]



import logging
import time
from asyncio import Event
from typing import Union, List, Any, Optional

from pyobs.interfaces import IMotion
from pyobs.modules import Module
from pyobs.utils.enums import MotionStatus
from pyobs.utils.parallel import event_wait

log = logging.getLogger(__name__)


class WaitForMotionMixin:
    """Mixin for a device that should wait for the motion status of another device."""

    __module__ = "pyobs.mixins"

    def __init__(
        self,
        wait_for_modules: Optional[List[str]] = None,
        wait_for_states: Optional[List[Union[MotionStatus, str]]] = None,
        wait_for_timeout: float = 0,
        **kwargs: Any,
    ):
        """Initializes the mixin.

        Args:
            wait_for_modules: One or more modules to wait for.
            wait_for_states: List of states to wait for.
            wait_for_timeout: Wait timeout in seconds.
        """

        # store
        self.__wait_for_modules = wait_for_modules if wait_for_modules is not None else []
        self.__wait_for_states = (
            [s if isinstance(s, MotionStatus) else MotionStatus(s) for s in wait_for_states]
            if wait_for_states is not None
            else []
        )
        self.__wait_for_timeout = wait_for_timeout

    async def _wait_for_motion(self, abort: Event) -> None:
        """Wait until all devices are in one of the given motion states.

        Args:
            abort: Abort event.

        Raises:
            TimeoutError: If wait timed out.
        """

        # no device?
        if len(self.__wait_for_modules) == 0:
            return

        # check type
        this = self
        if not isinstance(self, Module):
            raise ValueError("This is not a module.")

        # get all proxies
        proxies = [await self.proxy(device) for device in this.__wait_for_modules]

        # all need to be derived from IMotion
        if not all([isinstance(p, IMotion) for p in proxies]):
            raise ValueError("Not all given devices are derived from IMotion!")

        # run until timeout
        start = time.time()
        log.info("Waiting for motion of other modules...")
        while not abort.is_set():
            # timeout?
            if time.time() > start + this.__wait_for_timeout:
                raise TimeoutError

            # get all states and compare them
            states = [await p.get_motion_status() for p in proxies]

            # in a good state?
            good = [s in this.__wait_for_states for s in states]

            # if all good, we're finished waiting
            if all(good):
                log.info("All other modules have finished moving.")
                break

            # sleep a little
            await event_wait(abort, 1)


__all__ = ["WaitForMotionMixin"]



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011 Nathanael C. Fritz, Dann Martens (TOMOTON).
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from typing import Any, Dict, List, Optional
from slixmpp.xmlstream import ET
import base64
import logging
import time

log = logging.getLogger(__name__)

_namespace = "jabber:iq:rpc"


def fault2xml(code: int, message: str) -> ET.Element:
    value: Dict[str, Any] = dict()
    value["faultCode"] = code
    value["faultString"] = message
    fault = ET.Element("fault", {"xmlns": _namespace})
    fault.append(_py2xml((value,)))
    return fault


def xml2fault(params: ET.Element) -> Dict[str, Any]:
    vals = []
    for value in params.findall("{%s}value" % _namespace):
        vals.append(_xml2py(value)[0])
    fault = dict()
    fault["code"] = vals[0]["faultCode"]
    fault["string"] = vals[0]["faultString"]
    return fault


def py2xml(*args: Any) -> ET.Element:
    params = ET.Element("{%s}params" % _namespace)
    for x in args:
        param = ET.Element("{%s}param" % _namespace)
        param.append(_py2xml(x))
        params.append(param)  # <params><param>...
    return params


def _py2xml(*args: Any) -> ET.Element:
    for x in args:
        val = ET.Element("{%s}value" % _namespace)
        if x is None:
            nil = ET.Element("{%s}nil" % _namespace)
            val.append(nil)
        elif type(x) is int:
            i4 = ET.Element("{%s}i4" % _namespace)
            i4.text = str(x)
            val.append(i4)
        elif type(x) is bool:
            boolean = ET.Element("{%s}boolean" % _namespace)
            boolean.text = str(int(x))
            val.append(boolean)
        elif type(x) is str:
            string = ET.Element("{%s}string" % _namespace)
            string.text = x
            val.append(string)
        elif type(x) is float:
            double = ET.Element("{%s}double" % _namespace)
            double.text = str(x)
            val.append(double)
        elif type(x) is rpcbase64:
            b64 = ET.Element("{%s}base64" % _namespace)
            b64.text = x.encoded()
            val.append(b64)
        elif type(x) is rpctime:
            iso = ET.Element("{%s}dateTime.iso8601" % _namespace)
            iso.text = str(x)
            val.append(iso)
        elif type(x) in (list, tuple):
            array = ET.Element("{%s}array" % _namespace)
            data = ET.Element("{%s}data" % _namespace)
            for y in x:
                data.append(_py2xml(y))
            array.append(data)
            val.append(array)
        elif type(x) is dict:
            struct = ET.Element("{%s}struct" % _namespace)
            for y in x.keys():
                member = ET.Element("{%s}member" % _namespace)
                name = ET.Element("{%s}name" % _namespace)
                name.text = y
                member.append(name)
                member.append(_py2xml(x[y]))
                struct.append(member)
            val.append(struct)
        return val


def xml2py(params: ET.Element) -> List[Any]:
    namespace = "jabber:iq:rpc"
    vals: List[Any] = []
    for param in params.findall("{%s}param" % namespace):
        vals.append(_xml2py(param.find("{%s}value" % namespace)))
    return vals


def _xml2py(value: ET) -> Any:
    namespace = "jabber:iq:rpc"
    find_value = value.find
    if find_value("{%s}nil" % namespace) is not None:
        return None
    if find_value("{%s}i4" % namespace) is not None:
        return int(find_value("{%s}i4" % namespace).text)
    if find_value("{%s}int" % namespace) is not None:
        return int(find_value("{%s}int" % namespace).text)
    if find_value("{%s}boolean" % namespace) is not None:
        return bool(int(find_value("{%s}boolean" % namespace).text))
    if find_value("{%s}string" % namespace) is not None:
        return find_value("{%s}string" % namespace).text
    if find_value("{%s}double" % namespace) is not None:
        return float(find_value("{%s}double" % namespace).text)
    if find_value("{%s}base64" % namespace) is not None:
        return rpcbase64(find_value("{%s}base64" % namespace).text.encode())
    if find_value("{%s}Base64" % namespace) is not None:
        # Older versions of XEP-0009 used Base64
        return rpcbase64(find_value("{%s}Base64" % namespace).text.encode())
    if find_value("{%s}dateTime.iso8601" % namespace) is not None:
        return rpctime(find_value("{%s}dateTime.iso8601" % namespace).text)
    if find_value("{%s}struct" % namespace) is not None:
        struct = {}
        for member in find_value("{%s}struct" % namespace).findall("{%s}member" % namespace):
            struct[member.find("{%s}name" % namespace).text] = _xml2py(member.find("{%s}value" % namespace))
        return struct
    if find_value("{%s}array" % namespace) is not None:
        array = []
        for val in find_value("{%s}array" % namespace).find("{%s}data" % namespace).findall("{%s}value" % namespace):
            array.append(_xml2py(val))
        return array
    raise ValueError()


class rpcbase64(object):
    def __init__(self, data: bytes):
        # base 64 encoded string
        self.data = data

    def decode(self) -> bytes:
        return base64.b64decode(self.data)

    def __str__(self) -> str:
        return self.decode().decode()

    def encoded(self) -> str:
        return self.data.decode()


class rpctime(object):
    def __init__(self, data: Optional[Any] = None):
        # assume string data is in iso format YYYYMMDDTHH:MM:SS
        if type(data) is str:
            self.timestamp = time.strptime(data, "%Y%m%dT%H:%M:%S")
        elif type(data) is time.struct_time:
            self.timestamp = data
        elif data is None:
            self.timestamp = time.gmtime()
        else:
            raise ValueError()

    def iso8601(self) -> str:
        # return a iso8601 string
        return time.strftime("%Y%m%dT%H:%M:%S", self.timestamp)

    def __str__(self) -> str:
        return self.iso8601()



import time
from collections import namedtuple
from threading import Lock
from typing import Any, Dict


class DataCacheEntry:
    """A single entry in the data cache."""

    def __init__(self, name: str, data: Any = None):
        """Create a new entry for the data cache

        Args:
            name: Name of item
            data: Data for this item or None.
        """
        self.name = name
        self._data = data
        self.time = time.time()

    def update(self) -> None:
        """Update time for this item."""
        self.time = time.time()

    @property
    def data(self) -> Any:
        """Update usage time and return data for entry."""

        # update time
        self.update()

        # return data
        return self._data


class DataCache(object):
    """Data cache for proxy server."""

    """Type for cache entries."""
    Entry = namedtuple("Entry", "time name data")

    def __init__(self, size: int = 20):
        """Init cache.

        Args:
            size: Cache size.
        """
        self._lock = Lock()
        self._cache: Dict[str, DataCacheEntry] = {}
        self._size = size

    def __contains__(self, name: str) -> bool:
        """Checks, whether entry is in cache.

        Args:
            name: Name of entry.

        Returns:
            Whether it exists in cache.
        """
        with self._lock:
            return name in self._cache

    def __getitem__(self, name: str) -> Any:
        """Returns data from entry in cache.

        Args:
            name: Name of data.

        Returns:
            Data from entry in cache.

        Raises:
            IndexError: If entry does not exists.
        """

        # return data or raise IndexError
        with self._lock:
            return self._cache[name].data

    def __setitem__(self, name: str, data: Any) -> None:
        """Set new entry in the cache.

        Args:
            name: Name for data to store.
            data: Date of file.
        """

        # lock cache
        with self._lock:
            # does it exist already?
            if name in self._cache:
                # delete it
                del self._cache[name]

            # create new entry
            self._cache[name] = DataCacheEntry(name, data)

            # check size
            if len(self._cache) > self._size:
                # sort cache values by update time
                cache = sorted(self._cache.values(), key=lambda x: x.time, reverse=True)

                # delete all elements except for the latest N
                for c in cache[self._size :]:
                    del self._cache[c.name]

    def __delitem__(self, name: str) -> None:
        """Delete entry in cache.

        Args:
            name: Name of entry to delete.
        """
        del self._cache[name]


__all__ = ["DataCache"]



import asyncio
from functools import partial
from typing import Union, Optional, List, Tuple, Any, Dict, cast
import logging
import astropy.units as u

from pyobs.images.processor import ImageProcessor
from pyobs.images import Image
from pyobs.object import get_object
from pyobs.utils.archive import Archive
from pyobs.utils.enums import ImageType
from pyobs.utils.pipeline import Pipeline
from pyobs.utils.time import Time


log = logging.getLogger(__name__)


class Calibration(ImageProcessor):
    """Calibrate an image."""

    __module__ = "pyobs.images.processors.misc"

    """Cache for calibration frames."""
    calib_cache: List[Tuple[Tuple[ImageType, str, str, Optional[str]], Image]] = []

    def __init__(
        self,
        archive: Union[Dict[str, Any], Archive],
        max_cache_size: int = 20,
        max_days_bias: Optional[float] = None,
        max_days_dark: Optional[float] = None,
        max_days_flat: Optional[float] = None,
        **kwargs: Any,
    ):
        """Init a new image calibration pipeline step.

        Args:
            archive: Archive to fetch calibration frames from.
        """
        ImageProcessor.__init__(self, **kwargs)

        # store
        self._max_cache_size = max_cache_size
        self._max_days_bias = max_days_bias
        self._max_days_dark = max_days_dark
        self._max_days_flat = max_days_flat

        # get archive
        self._archive = get_object(archive, Archive)

    async def __call__(self, image: Image) -> Image:
        """Calibrate an image.

        Args:
            image: Image to calibrate.

        Returns:
            Calibrated image.
        """
        import ccdproc

        # get calibration masters
        try:
            bias = await self._find_master(image, ImageType.BIAS, max_days=self._max_days_bias)
            dark = await self._find_master(image, ImageType.DARK, max_days=self._max_days_dark)
            flat = await self._find_master(image, ImageType.SKYFLAT, max_days=self._max_days_flat)
        except ValueError as e:
            log.error("Could not find calibration frames: " + str(e))
            return image

        # calibrate image
        c = await asyncio.get_running_loop().run_in_executor(
            None,
            partial(
                ccdproc.ccd_process,
                image.to_ccddata(),
                oscan=image.header["BIASSEC"] if "BIASSEC" in image.header else None,
                trim=image.header["TRIMSEC"] if "TRIMSEC" in image.header else None,
                error=True,
                master_bias=bias.to_ccddata() if bias is not None else None,
                dark_frame=dark.to_ccddata() if dark is not None else None,
                master_flat=flat.to_ccddata() if flat is not None else None,
                bad_pixel_mask=None,
                gain=image.header["DET-GAIN"] * u.electron / u.adu,
                readnoise=image.header["DET-RON"] * u.electron,
                dark_exposure=dark.header["EXPTIME"] * u.second if dark is not None else None,
                data_exposure=image.header["EXPTIME"] * u.second,
                dark_scale=True,
                gain_corrected=False,
            ),
        )

        # to image
        calibrated = Image.from_ccddata(c)
        calibrated.header["BUNIT"] = ("electron", "Unit of pixel values")

        # set raw filename
        if "ORIGNAME" in image.header:
            calibrated.header["L1RAW"] = image.header["ORIGNAME"].replace(".fits", "")

        # add calibration frames
        if bias is not None:
            calibrated.header["L1BIAS"] = (
                bias.header["FNAME"].replace(".fits.fz", "").replace(".fits", ""),
                "Name of BIAS frame",
            )
        if dark is not None:
            calibrated.header["L1DARK"] = (
                dark.header["FNAME"].replace(".fits.fz", "").replace(".fits", ""),
                "Name of DARK frame",
            )
        if flat is not None:
            calibrated.header["L1FLAT"] = (
                flat.header["FNAME"].replace(".fits.fz", "").replace(".fits", ""),
                "Name of FLAT frame",
            )

        # set RLEVEL
        calibrated.header["RLEVEL"] = (1, "Reduction level")

        # finished
        return calibrated

    async def _find_master(
        self, image: Image, image_type: ImageType, max_days: Optional[float] = None
    ) -> Optional[Image]:
        """Find master calibration frame for given parameters using a cache.

        Args:
            image_type: image type.

        Returns:
            Image or None

        Raises:
            ValueError: if no calibration frame could be found.
        """

        # get mode
        try:
            instrument = image.header["INSTRUME"]
            binning = "{0}x{0}".format(image.header["XBINNING"])
            filter_name = cast(str, image.header["FILTER"]) if "FILTER" in image.header else None
            time = Time(image.header["DATE-OBS"])
            mode = image_type, instrument, binning, filter_name
        except KeyError:
            # could not fetch header items
            raise ValueError("Could not fetch items from image header.")

        # is in cache?
        for m, item in Calibration.calib_cache:
            if m == mode:
                return item

        # try to download one
        master = await Pipeline.find_master(
            self._archive,
            image_type,
            time,
            instrument,
            binning,
            None if image_type in [ImageType.BIAS, ImageType.DARK] else filter_name,
            max_days=max_days,
        )

        # nothing?
        if master is None:
            raise ValueError("No master frame found.")

        # store it in cache
        Calibration.calib_cache.append((mode, master))

        # too many entries?
        while len(Calibration.calib_cache) > self._max_cache_size:
            Calibration.calib_cache.pop(0)

        # return it
        return master


__all__ = ["Calibration"]



"""
Mixins are classes that can be inherited from to automatically add some functionality to a module.
"""
__title__ = "Mixins"

from .camerasettings import CameraSettingsMixin
from .fitsnamespace import FitsNamespaceMixin
from .follow import FollowMixin
from .motionstatus import MotionStatusMixin
from .waitformotion import WaitForMotionMixin
from .weatheraware import WeatherAwareMixin
from .fitsheader import FitsHeaderMixin, ImageFitsHeaderMixin, SpectrumFitsHeaderMixin



from abc import ABCMeta, abstractmethod
from typing import Any

from .interface import Interface


class IFocusModel(Interface, metaclass=ABCMeta):
    """The module provides a model for the telescope focus, e.g. based on temperatures."""

    __module__ = "pyobs.interfaces"

    @abstractmethod
    async def get_optimal_focus(self, **kwargs: Any) -> float:
        """Returns the optimal focus."""
        ...

    @abstractmethod
    async def set_optimal_focus(self, **kwargs: Any) -> None:
        """Sets optimal focus."""
        ...


__all__ = ["IFocusModel"]



import asyncio
from typing import Tuple, Any
import logging

from .sourcedetection import SourceDetection
from pyobs.images import Image

log = logging.getLogger(__name__)


class DaophotSourceDetection(SourceDetection):
    """Detect source using Daophot."""

    __module__ = "pyobs.images.processors.detection"

    def __init__(
        self,
        fwhm: float = 3.0,
        threshold: float = 4.0,
        bkg_sigma: float = 3.0,
        bkg_box_size: Tuple[int, int] = (50, 50),
        bkg_filter_size: Tuple[int, int] = (3, 3),
        **kwargs: Any,
    ):
        """Initializes a wrapper for photutils. See its documentation for details.

        Args:
            fwhm: Full-width at half maximum for Gaussian kernel.
            threshold: Threshold pixel value for detection.
            bkg_sigma: Sigma for background kappa-sigma clipping.
            bkg_box_size: Box size for background estimation.
            bkg_filter_size: Filter size for background estimation.
        """
        SourceDetection.__init__(self, **kwargs)

        # store
        self.fwhm = fwhm
        self.threshold = threshold
        self.bkg_sigma = bkg_sigma
        self.bkg_box_size = bkg_box_size
        self.bkg_filter_size = bkg_filter_size

    async def __call__(self, image: Image) -> Image:
        """Find stars in given image and append catalog.

        Args:
            image: Image to find stars in.

        Returns:
            Image with attached catalog.
        """
        from astropy.stats import SigmaClip, sigma_clipped_stats
        from photutils import Background2D, MedianBackground, DAOStarFinder

        # get data
        if image.data is None:
            log.warning("No data found in image.")
            return image
        data = image.data.astype(float).copy()

        # estimate background
        sigma_clip = SigmaClip(sigma=self.bkg_sigma)
        bkg_estimator = MedianBackground()
        bkg = Background2D(
            data,
            self.bkg_box_size,
            filter_size=self.bkg_filter_size,
            sigma_clip=sigma_clip,
            bkg_estimator=bkg_estimator,
            mask=image.mask,
        )
        data -= bkg.background

        # do statistics
        mean, median, std = sigma_clipped_stats(data, sigma=3.0)

        # find stars
        daofind = DAOStarFinder(fwhm=self.fwhm, threshold=self.threshold * std)
        loop = asyncio.get_running_loop()
        sources = await loop.run_in_executor(None, daofind, data - median)

        # rename columns
        sources.rename_column("xcentroid", "x")
        sources.rename_column("ycentroid", "y")

        # match fits conventions
        sources["x"] += 1
        sources["y"] += 1

        # pick columns for catalog
        cat = sources["x", "y", "flux", "peak"]

        # copy image, set catalog and return it
        img = image.copy()
        img.catalog = cat
        return img


__all__ = ["DaophotSourceDetection"]



from __future__ import annotations

import asyncio
import functools
import inspect
import json
import logging
import re
import ssl
import time
from collections import Coroutine
from typing import Any, Callable, Dict, Type, List, Optional, TYPE_CHECKING
from sleekxmpp import ElementBase
from sleekxmpp.xmlstream import ET
import xml.sax.saxutils

from pyobs.comm import Comm
from pyobs.events import Event, LogEvent, ModuleOpenedEvent, ModuleClosedEvent
from pyobs.events.event import EventFactory
from .rpc import RPC
from .xmppclient import XmppClient
from ...interfaces import Interface
from ...utils.parallel import Future

if TYPE_CHECKING:
    from pyobs.modules import Module

log = logging.getLogger(__name__)


class EventStanza(ElementBase):  # type: ignore
    name = "event"
    namespace = "pyobs:event"


class XmppComm(Comm):
    """A Comm class using XMPP.

    This Comm class uses an XMPP server (e.g. `ejabberd <https://www.ejabberd.im>`_) for communication between modules.
    Essentially required for a connection to the server is a JID, a JabberID. It can be specified in the configuration
    like this::

        comm:
            class: pyobs.sleekxmpp.XmppComm
            jid:  someuser@example.com/pyobs

    Using this, *pyobs* tries to connect to example.com as user ``someuser`` with resource ``pyobs``. Since ``pyobs``
    is the default resource, it can be omitted::

        jid:  someuser@example.com

    Alternatively, one can split the user, domain, and resource (if required) into three different parameters::

        user: someuser
        domain: example.com

    This comes in handy, if one wants to put the basic Comm configuration into a separate file. Imagine a ``_comm.yaml``
    in the same directory as the module config::

        comm_cfg: &comm
            class: pyobs.comm.sleekxmpp.XmppComm
            domain: example.com

    Now in the module configuration, one can simply do this::

        {include _comm.yaml}

        comm:
            <<: *comm
            user: someuser
            password: supersecret

    This allows for a super easy change of the domain for all configurations, which especially makes developing on
    different machines a lot easier.

    The ``server`` parameter can be used, when the server's hostname is different from the XMPP domain. This might,
    e.g., be the case, when connecting to a server via SSH port forwarding::

        jid:  someuser@example.com/pyobs
        server: localhost:52222

    Finally, always make sure that ``use_tls`` is set according to the server's settings, i.e. if it uses TLS, this
    parameter must be True, and False otherwise. Cryptic error messages will follow, if one does not set this properly.

    """

    __module__ = "pyobs.comm.sleekxmpp"

    def __init__(
        self,
        jid: Optional[str] = None,
        user: Optional[str] = None,
        domain: Optional[str] = None,
        resource: str = "pyobs",
        password: str = "",
        server: Optional[str] = None,
        use_tls: bool = False,
        *args: Any,
        **kwargs: Any,
    ):
        """Create a new XMPP Comm module.

        Either a fill JID needs to be provided, or a set of user/domian/resource, from which a JID is built.

        Args:
            jid: JID to connect as.
            user: Username part of the JID.
            domain: Domain part of the JID.
            resource: Resource part of the JID.
            password: Password for given JID.
            server: Server to connect to. If not given, domain from JID is used.
            use_tls: Whether or not to use TLS.
        """
        Comm.__init__(self, *args, **kwargs)

        # variables
        self._rpc: Optional[RPC] = None
        self._connected = False
        self._event_handlers: Dict[Type[Event], List[Callable[[Event, str], Coroutine[Any, Any, bool]]]] = {}
        self._online_clients: List[str] = []
        self._interface_cache: Dict[str, List[Type[Interface]]] = {}
        self._user = user
        self._domain = domain
        self._resource = resource
        self._server = server
        self._use_tls = use_tls
        self._loop = None

        # build jid
        if jid:
            # resource given in jid?
            if "/" not in jid:
                jid += "/" + resource

            # get user/domain/resource and write it back to config
            m = re.match(r"([\w_\-\.]+)@([\w_\-\.]+)\/([\w_\-\.]+)", jid)
            if not m:
                log.error("Invalid JID format.")
                raise ValueError()
            self._user = m.group(1)
            self._domain = m.group(2)
            self._resource = m.group(3)

            # set jid itself
            self._jid = jid

        else:
            self._jid = "%s@%s/%s" % (self._user, self._domain, self._resource)

        # create client
        self._xmpp = XmppClient(self._jid, password)
        self._xmpp.add_event_handler("pubsub_publish", self._handle_event)
        self._xmpp.add_event_handler("got_online", self._got_online)
        self._xmpp.add_event_handler("got_offline", self._got_offline)

    def _set_module(self, module: "Module") -> None:
        """Called, when the module connected to this Comm changes.

        Args:
            module: The module.
        """

        # add features
        if module is not None:
            for i in module.interfaces:
                self._xmpp["xep_0030"].add_feature("pyobs:interface:%s" % i.__name__)

        # update RPC
        if self._rpc is None:
            raise ValueError("No RPC.")
        self._rpc.set_handler(module)

    async def open(self) -> None:
        """Open the connection to the XMPP server.

        Returns:
            Whether opening was successful.
        """
        await Comm.open(self)

        # store loop
        self._loop = asyncio.get_running_loop()

        # create RPC handler
        self._rpc = RPC(self._xmpp, self._loop, self.module)

        # server given?
        server = () if self._server is None else tuple(self._server.split(":"))

        # connect
        self._xmpp.ssl_version = ssl.PROTOCOL_TLSv1_2
        if self._xmpp.connect(server, use_tls=self._use_tls, reattempt=False):
            # start processing
            self._xmpp.process(block=False)

            # wait for connected
            if not self._xmpp.wait_connect():
                raise ValueError("Could not connect to XMPP server.")
            self._connected = True

            # subscribe to events
            await self.register_event(LogEvent)

        else:
            # TODO: catch exceptions in open() methods
            raise ValueError("Could not connect to XMPP server.")

    async def close(self) -> None:
        """Close connection."""

        # close parent class
        await Comm.close(self)

        # disconnect from sleekxmpp server
        self._xmpp.disconnect()

    @property
    def name(self) -> Optional[str]:
        """Name of this client."""
        return self._user

    def _failed_auth(self, event: Any) -> None:
        """Authentification failed.

        Args:
            event: XMPP event.
        """
        print("Authorization at server failed.")

    def _get_full_client_name(self, name: str) -> str:
        """Builds full JID from a given username.

        Args:
            name: Username to build JID for.

        Returns:
            Full JID for given user.
        """
        return name if "@" in name else "%s@%s/%s" % (name, self._domain, self._resource)

    async def get_interfaces(self, client: str) -> List[Type[Interface]]:
        """Returns list of interfaces for given client.

        Args:
            client: Name of client.

        Returns:
            List of supported interfaces.

        Raises:
            IndexError: If client cannot be found.
        """

        # full JID given?
        if "@" not in client:
            client = "%s@%s/%s" % (client, self._domain, self._resource)

        # get interfaces
        if client not in self._interface_cache:
            # get interface names
            loop = asyncio.get_running_loop()
            interface_names = await loop.run_in_executor(None, self._xmpp.get_interfaces, client)

            # get interfaces
            self._interface_cache[client] = self._interface_names_to_classes(interface_names)

        # convert to classes
        return self._interface_cache[client]

    async def _supports_interface(self, client: str, interface: Type[Interface]) -> bool:
        """Checks, whether the given client supports the given interface.

        Args:
            client: Client to check.
            interface: Interface to check.

        Returns:
            Whether or not interface is supported.
        """

        # full JID given?
        if "@" not in client:
            client = "%s@%s/%s" % (client, self._domain, self._resource)

        # update interface cache and get interface names
        interfaces = await self.get_interfaces(client)

        # supported?
        return interface in interfaces

    async def execute(self, client: str, method: str, signature: inspect.Signature, *args: Any) -> Any:
        """Execute a given method on a remote client.

        Args:
            client (str): ID of client.
            method (str): Method to call.
            signature: Method signature.
            *args: List of parameters for given method.

        Returns:
            Passes through return from method call.
        """
        if self._rpc is None:
            raise ValueError("No RPC.")
        return await self._rpc.call(self._get_full_client_name(client), method, signature, *args)

    def _got_online(self, msg: Any) -> None:
        """If a new client connects, add it to list.

        Args:
            msg: XMPP message.
        """

        # append to list
        jid = msg["from"].full
        if jid not in self._online_clients:
            self._online_clients.append(jid)

        # clear interface cache, just in case there is something there
        if jid in self._interface_cache:
            del self._interface_cache[jid]

        # send event
        self._send_event_to_module(ModuleOpenedEvent(), msg["from"].username)

    def _got_offline(self, msg: Any) -> None:
        """If a new client disconnects, remove it from list.

        Args:
            msg: XMPP message.
        """

        # remove from list
        jid = msg["from"].full
        self._online_clients.remove(jid)

        # clear interface cache
        if jid in self._interface_cache:
            del self._interface_cache[jid]

        # send event
        self._send_event_to_module(ModuleClosedEvent(), msg["from"].username)

    @property
    def clients(self) -> List[str]:
        """Returns list of currently connected clients.

        Returns:
            (list) List of currently connected clients.
        """
        return [c[: c.find("@")] for c in self._online_clients]

    @property
    def client(self) -> XmppClient:
        """Returns the XMPP client.

        Returns:
            The XMPP client.
        """
        return self._xmpp

    async def send_event(self, event: Event) -> None:
        """Send an event to other clients.

        Args:
            event (Event): Event to send
        """

        # create stanza
        stanza = EventStanza()

        # dump event to JSON and escape it
        body = xml.sax.saxutils.escape(json.dumps(event.to_json()))

        # set xml and send event
        stanza.xml = ET.fromstring('<event xmlns="pyobs:event">%s</event>' % body)
        self._xmpp["xep_0163"].publish(
            stanza,
            node="pyobs:event:%s" % event.__class__.__name__,
            block=False,
            callback=functools.partial(self._send_event_callback, event=event),
        )

    @staticmethod
    def _send_event_callback(iq: Any, event: Optional[Event] = None) -> None:
        """Called when an event has been successfully sent.

        Args:
            iq: Response package.
            event: Sent event.
        """
        log.debug("%s successfully sent.", event.__class__.__name__)

    async def register_event(
        self, event_class: Type[Event], handler: Optional[Callable[[Event, str], Coroutine[Any, Any, bool]]] = None
    ) -> None:
        """Register an event type. If a handler is given, we also receive those events, otherwise we just
        send them.

        Args:
            event_class: Class of event to register.
            handler: Event handler method.
        """

        # we also want to register all events derived from the given one
        event_classes = self._get_derived_events(event_class)

        # do we have a handler?
        if handler:
            # loop classes
            for ev in event_classes:
                # initialize list
                if ev not in self._event_handlers:
                    self._event_handlers[ev] = []
                # avoid duplicates
                if handler not in self._event_handlers[ev]:
                    # add handler
                    self._event_handlers[ev].append(handler)

        # if event is not a local one, we also need to do some XMPP stuff
        if not event_class.local:
            self._register_events(event_classes, handler)

    def _get_derived_events(self, event: Type[Event]) -> List[Type[Event]]:
        """Return list of given event itself and all events derived from it.

        Args:
            event: Event class to check.

        Returns:
            List of event classes.
        """
        import pyobs.events

        event_classes: List[Type[Event]] = []
        for cls in inspect.getmembers(pyobs.events, inspect.isclass):
            if issubclass(cls[1], event):
                event_classes.append(cls[1])
        return event_classes

    def _register_events(
        self, events: List[Type[Event]], handler: Optional[Callable[[Event, str], bool]] = None
    ) -> None:
        # loop events
        for ev in events:
            # register event at XMPP
            self._xmpp["xep_0030"].add_feature("pyobs:event:%s" % ev.__name__)

            # if we have a handler, we're also interested in receiving such events
            if handler:
                # add interest
                self._xmpp["xep_0163"].add_interest("pyobs:event:%s" % ev.__name__)

        # update caps and send presence
        self._xmpp["xep_0115"].update_caps()
        self._xmpp.send_presence()

    def _handle_event(self, msg: Any) -> None:
        """Handles an event.

        Args:
            msg: Received XMPP message.
        """

        # get body, unescape it, parse it
        # node = msg['pubsub_event'][items']['node']
        body = json.loads(xml.sax.saxutils.unescape(msg["pubsub_event"]["items"]["item"]["payload"].text))

        # do we have a <delay> element?
        delay = msg.findall("{urn:sleekxmpp:delay}delay")
        if len(delay) > 0:
            # ignore this message
            return

        # did we send this?
        if msg["from"] == self._xmpp.boundjid.bare:
            return

        # create event and check timestamp
        event = EventFactory.from_dict(body)
        if event is None:
            return
        if time.time() - event.timestamp > 30:
            # event is more than 30 seconds old, ignore it
            # we do this do avoid resent events after a reconnect
            return

        # send it to module
        self._send_event_to_module(event, msg["from"].username)

    def _send_event_to_module(self, event: Event, from_client: str) -> None:
        """Send an event to all connected modules.

        Args:
            event: Event to send.
            from_client: Client that sent the event.
        """

        # send it
        if event.__class__ in self._event_handlers:
            for handler in self._event_handlers[event.__class__]:
                # run handler
                if asyncio.iscoroutinefunction(handler):
                    asyncio.run_coroutine_threadsafe(handler(event, from_client), self._loop)
                else:
                    handler(event, from_client)


__all__ = ["XmppComm"]



import logging
from typing import Union, Tuple, Dict, Any, Optional
import threading
import numpy as np
from numpy.typing import NDArray

from pyobs.interfaces import IAutoFocus
from pyobs.events import FocusFoundEvent
from pyobs.interfaces import IExposureTime, IImageType, IFocuser, IFilters, IImageGrabber
from pyobs.object import get_object
from pyobs.mixins import CameraSettingsMixin
from pyobs.modules import timeout, Module
from pyobs.utils.enums import ImageType
from pyobs.utils.focusseries import FocusSeries
from pyobs.utils import exceptions as exc

log = logging.getLogger(__name__)


class AutoFocusSeries(Module, CameraSettingsMixin, IAutoFocus):
    """Module for auto-focusing a telescope."""

    __module__ = "pyobs.modules.focus"

    def __init__(
        self,
        focuser: Union[str, IFocuser],
        camera: Union[str, IImageGrabber],
        series: Union[Dict[str, Any], FocusSeries],
        offset: bool = False,
        filters: Optional[Union[str, IFilters]] = None,
        filter_name: Optional[str] = None,
        binning: Optional[int] = None,
        **kwargs: Any,
    ):
        """Initialize a new auto focus system.

        Args:
            focuser: Name of IFocuser.
            camera: Name of ICamera.
            filters: Name of IFilters, if any.
            filter_name: Name of filter to set.
            offset: If True, offsets are used instead of absolute focus values.
        """
        Module.__init__(self, **kwargs)

        # store focuser and camera
        self._focuser = focuser
        self._camera = camera
        self._filters = filters
        self._offset = offset
        self._abort = threading.Event()

        # create focus series
        self._series: FocusSeries = get_object(series, FocusSeries)

        # init camera settings mixin
        CameraSettingsMixin.__init__(self, filters=filters, filter_name=filter_name, binning=binning, **kwargs)

        # register exceptions
        if isinstance(camera, str):
            exc.register_exception(
                exc.RemoteError, 3, timespan=600, module=camera, callback=self._default_remote_error_callback
            )
        if isinstance(focuser, str):
            exc.register_exception(
                exc.RemoteError, 3, timespan=600, module=focuser, callback=self._default_remote_error_callback
            )

    async def open(self) -> None:
        """Open module"""
        await Module.open(self)

        # register event
        await self.comm.register_event(FocusFoundEvent)

        # check focuser and camera
        try:
            await self.proxy(self._focuser, IFocuser)
            await self.proxy(self._camera, IImageGrabber)
        except ValueError:
            log.warning("Either camera or focuser do not exist or are not of correct type at the moment.")

    @timeout(600)
    async def auto_focus(self, count: int, step: float, exposure_time: float, **kwargs: Any) -> Tuple[float, float]:
        """Perform an auto-focus series.

        This method performs an auto-focus series with "count" images on each side of the initial guess and the given
        step size. With count=3, step=1 and guess=10, this takes images at the following focus values:
        7, 8, 9, 10, 11, 12, 13

        Args:
            count: Number of images to take on each side of the initial guess. Should be an odd number.
            step: Step size.
            exposure_time: Exposure time for images.

        Returns:
            Tuple of obtained best focus value and its uncertainty. Or Nones, if focus series failed.

        Raises:
            FileNotFoundException: If image could not be downloaded.
        """
        log.info("Performing auto-focus...")

        # get focuser
        log.info("Getting proxy for focuser...")
        focuser = await self.proxy(self._focuser, IFocuser)

        # get camera
        log.info("Getting proxy for camera...")
        camera = await self.proxy(self._camera, IImageGrabber)

        # do camera settings
        await self._do_camera_settings(camera)

        # get filter wheel and current filter
        filter_name = "unknown"
        try:
            filter_wheel = await self.proxy(self._filters, IFilters)
            filter_name = await filter_wheel.get_filter()
        except ValueError:
            log.warning("Filter module is not of type IFilters. Could not get filter.")

        # get focus as first guess
        try:
            if self._offset:
                guess = 0.0
                log.info("Using focus offset of 0mm as initial guess.")
            else:
                guess = await focuser.get_focus()
                log.info("Using current focus of %.2fmm as initial guess.", guess)
        except exc.RemoteError:
            raise ValueError("Could not fetch current focus value.")

        # define array of focus values to iterate
        focus_values: NDArray[float] = np.linspace(guess - count * step, guess + count * step, 2 * count + 1)

        # reset
        self._series.reset()
        self._abort = threading.Event()

        # loop focus values
        log.info("Starting focus series...")
        for foc in focus_values:
            # set focus
            if self._offset:
                log.info("Changing focus offset to %.2fmm...", foc)
            else:
                log.info("Changing focus to %.2fmm...", foc)
            if self._abort.is_set():
                raise InterruptedError()
            try:
                if self._offset:
                    await focuser.set_focus_offset(float(foc))
                else:
                    await focuser.set_focus(float(foc))

            except exc.RemoteError:
                raise ValueError("Could not set new focus value.")

            # do exposure
            log.info("Taking picture...")
            if self._abort.is_set():
                raise InterruptedError()
            try:
                if isinstance(camera, IExposureTime):
                    await camera.set_exposure_time(exposure_time)
                if isinstance(camera, IImageType):
                    await camera.set_image_type(ImageType.FOCUS)
                filename = await camera.grab_image()
            except exc.RemoteError:
                log.error("Could not take image.")
                continue

            # download image
            log.info("Downloading image...")
            image = await self.vfs.read_image(filename)

            # analyse
            log.info("Analysing picture...")
            try:
                self._series.analyse_image(image, foc)
            except:
                # do nothing..
                log.error("Could not analyse image.")
                continue

        # fit focus
        if self._abort.is_set():
            raise InterruptedError()
        focus = self._series.fit_focus()

        # did focus series fail?
        if focus is None or focus[0] is None or np.isnan(focus[0]):
            log.warning("Focus series failed.")

            # reset to initial values
            if self._offset:
                log.info("Resetting focus offset to 0.", guess)
                await focuser.set_focus_offset(0)
            else:
                log.info("Resetting focus to initial guess of %.3f mm.", guess)
                await focuser.set_focus(guess)

            # raise error
            raise ValueError("Could not find best focus.")

        # log and set focus
        if self._offset:
            log.info("Setting new focus offset of (%.3f+-%.3f) mm.", focus[0], focus[1])
            absolute = focus[0] + await focuser.get_focus()
            await focuser.set_focus_offset(focus[0])
        else:
            log.info("Setting new focus value of (%.3f+-%.3f) mm.", focus[0], focus[1])
            absolute = focus[0] + await focuser.get_focus_offset()
            await focuser.set_focus(focus[0])

        # send event
        await self.comm.send_event(FocusFoundEvent(absolute, focus[1], filter_name))

        # return result
        return focus[0], focus[1]

    async def auto_focus_status(self, **kwargs: Any) -> Dict[str, Any]:
        """Returns current status of auto focus.

        Returned dictionary contains a list of focus/fwhm pairs in X and Y direction.

        Returns:
            Dictionary with current status.
        """
        return {}

    @timeout(20)
    async def abort(self, **kwargs: Any) -> None:
        """Abort current actions."""
        self._abort.set()


__all__ = ["AutoFocusSeries"]



import logging
from typing import Any


class DuplicateFilter(logging.Filter):
    """Logging filter that removes duplicate entries.

    Should be used with new logger, e.g.:
        log = logging.getLogger('filtered_logger')
        log.addFilter(DuplicateFilter())
        log.info('Test')
    """

    def __init__(self, *args: Any, **kwargs: Any):
        logging.Filter.__init__(self, *args, **kwargs)
        self.last_log = ""

    def filter(self, record: Any) -> bool:
        if record.getMessage() != getattr(self, "last_log", None):
            self.last_log = record.getMessage()
            return True
        return False


__all__ = ["DuplicateFilter"]



from abc import ABCMeta, abstractmethod
from typing import Any

from pyobs.object import Object


class Publisher(Object, metaclass=ABCMeta):
    @abstractmethod
    async def __call__(self, **kwargs: Any) -> None:
        """Publish the given results.

        Args:
            **kwargs: Results to publish.
        """
        ...


__all__ = ["Publisher"]



import os
from tempfile import NamedTemporaryFile
import logging
from typing import Optional, Any, AnyStr, cast

from .localfile import VFSFile


log = logging.getLogger(__name__)


class TempFile(VFSFile):
    """A temporary file."""

    __module__ = "pyobs.vfs"

    def __init__(
        self,
        name: Optional[str] = None,
        mode: str = "r",
        prefix: Optional[str] = None,
        suffix: Optional[str] = None,
        root: str = "/tmp/pyobs/",
        mkdir: bool = True,
        **kwargs: Any,
    ):
        """Open/create a temp file.

        Args:
            name: Name of file.
            mode: Open mode.
            prefix: Prefix for automatic filename creation in write mode.
            suffix: Suffix for automatic filename creation in write mode.
            root: Temp directory.
            mkdir: Whether to automatically create directories.
        """

        # no root given?
        if root is None:
            raise ValueError("No root directory given.")

        # create root?
        if not os.path.exists(root):
            os.makedirs(root)

        # no filename?
        if name is None:
            # cannot read from non-existing filename
            if "r" in mode:
                raise ValueError("No filename given to read from.")

            # create new temp file name
            with NamedTemporaryFile(mode=mode, prefix=prefix, suffix=suffix, dir=root) as tmp:
                name = os.path.basename(tmp.name)

        # filename is not allowed to start with a / or contain ..
        if name.startswith("/") or ".." in name:
            raise ValueError("Only files within root directory are allowed.")

        # build filename
        self.filename = name
        self.full_name = os.path.join(root, name)

        # need to create directory?
        path = os.path.dirname(self.full_name)
        if not os.path.exists(path):
            if mkdir:
                os.makedirs(path)
            else:
                raise ValueError("Cannot write into sub-directory with disabled mkdir option.")

        # init fd
        self.mode = mode
        self.fd = open(self.full_name, mode)

    async def close(self) -> None:
        """Close file."""

        # close file
        if self.fd:
            self.fd.close()

        # remove file
        if "w" in self.mode:
            os.remove(self.full_name)

    async def read(self, n: int = -1) -> AnyStr:
        if self.fd is None:
            raise OSError
        return cast(AnyStr, self.fd.read(n))

    async def write(self, s: AnyStr) -> None:
        if self.fd is None:
            raise OSError
        self.fd.write(s)


__all__ = ["TempFile"]
